from typing import List, Set, Dict, Optional
import stringcase
from cachetools import cached

from kgx.config import get_logger
from kgx.graph.base_graph import BaseGraph
from kgx.utils.kgx_utils import get_toolkit, get_cache, get_curie_lookup_service
from kgx.prefix_manager import PrefixManager

ONTOLOGY_PREFIX_MAP: Dict = {}
ONTOLOGY_GRAPH_CACHE: Dict = {}

log = get_logger()


def get_parents(graph: BaseGraph, node: str, relations: List[str] = None) -> List[str]:
    """
    Return all direct `parents` of a specified node, filtered by ``relations``.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        Graph to traverse
    node: str
        node identifier
    relations: List[str]
       list of relations

    Returns
    -------
    List[str]
        A list of parent node(s)

    """
    parents = []
    if graph.has_node(node):
        out_edges = [x for x in graph.out_edges(node, keys=False, data=True)]
        if relations is None:
            parents = [x[1] for x in out_edges]
        else:
            parents = [x[1] for x in out_edges if x[2]["predicate"] in relations]
    return parents


def get_ancestors(
    graph: BaseGraph, node: str, relations: List[str] = None
) -> List[str]:
    """
    Return all `ancestors` of specified node, filtered by ``relations``.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        Graph to traverse
    node: str
        node identifier
    relations: List[str]
       list of relations

    Returns
    -------
    List[str]
        A list of ancestor nodes

    """
    seen = []
    nextnodes = [node]
    while len(nextnodes) > 0:
        nn = nextnodes.pop()
        if nn not in seen:
            seen.append(nn)
            nextnodes += get_parents(graph, nn, relations=relations)
    seen.remove(node)
    return seen


@cached(get_cache())
def get_category_via_superclass(
    graph: BaseGraph, curie: str, load_ontology: bool = True
) -> Set[str]:
    """
    Get category for a given CURIE by tracing its superclass, via ``subclass_of`` hierarchy,
    and getting the most appropriate category based on the superclass.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        Graph to traverse
    curie: str
        Input CURIE
    load_ontology: bool
        Determines whether to load ontology, based on CURIE prefix, or to simply
        rely on ``subclass_of`` hierarchy from graph

    Returns
    -------
    Set[str]
        A set containing one (or more) category for the given CURIE

    """
    log.debug("curie: {}".format(curie))
    new_categories = []
    toolkit = get_toolkit()
    if PrefixManager.is_curie(curie):
        ancestors = get_ancestors(graph, curie, relations=["subclass_of"])
        if len(ancestors) == 0 and load_ontology:
            cls = get_curie_lookup_service()
            ontology_graph = cls.ontology_graph
            new_categories += [
                x for x in get_category_via_superclass(ontology_graph, curie, False)
            ]
        log.debug("Ancestors for CURIE {} via subClassOf: {}".format(curie, ancestors))
        seen = []
        for anc in ancestors:
            mapping = toolkit.get_by_mapping(anc)
            seen.append(anc)
            if mapping:
                # there is direct mapping to BioLink Model
                log.debug("Ancestor {} mapped to {}".format(anc, mapping))
                seen_labels = [
                    graph.nodes()[x]["name"] for x in seen if "name" in graph.nodes()[x]
                ]
                new_categories += [x for x in seen_labels]
                new_categories += [x for x in toolkit.ancestors(mapping)]
                break
    return set(new_categories)


def curie_lookup(curie: str) -> Optional[str]:
    """
    Given a CURIE, find its label.

    This method first does a lookup in predefined maps. If none found,
    it makes use of CurieLookupService to look for the CURIE in a set
    of preloaded ontologies.

    Parameters
    ----------
    curie: str
        A CURIE

    Returns
    -------
    Optional[str]
        The label corresponding to the given CURIE

    """
    cls = get_curie_lookup_service()
    name: Optional[str] = None
    prefix = PrefixManager.get_prefix(curie)
    if prefix in ["OIO", "OWL", "owl", "OBO", "rdfs"]:
        name = stringcase.snakecase(curie.split(":", 1)[1])
    elif curie in cls.curie_map:
        name = cls.curie_map[curie]
    elif curie in cls.ontology_graph:
        name = cls.ontology_graph.nodes()[curie]["name"]
    return name



from typing import Dict, Optional, List, Generator, Any


class BaseGraph(object):
    """
    BaseGraph that is a wrapper and provides methods to interact with a graph store.

    All implementations should extend this BaseGraph class and implement all the defined methods.
    """

    def __init__(self):
        self.graph = None
        self.name = None

    def add_node(self, node: str, **kwargs: Any) -> Any:
        """
        Add a node to the graph.

        Parameters
        ----------
        node: str
            Node identifier
        **kwargs: Any
            Any additional node properties

        """
        pass

    def add_edge(
        self,
        subject_node: str,
        object_node: str,
        edge_key: Optional[str] = None,
        **kwargs: Any
    ) -> Any:
        """
        Add an edge to the graph.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key
        kwargs: Any
            Any additional edge properties

        Returns
        -------
        Any

        """
        pass

    def add_node_attribute(self, node: str, key: str, value: Any) -> Any:
        """
        Add an attribute to a given node.

        Parameters
        ----------
        node: str
            The node identifier
        key: str
            The key for an attribute
        value: Any
            The value corresponding to the key

        Returns
        -------
        Any

        """
        pass

    def add_edge_attribute(
        self,
        subject_node: str,
        object_node: str,
        edge_key: Optional[str],
        attr_key: str,
        attr_value: Any,
    ) -> Any:
        """
        Add an attribute to a given edge.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key
        attr_key: str
            The attribute key
        attr_value: Any
            The attribute value

        Returns
        -------
        Any

        """
        pass

    def update_node_attribute(self, node, key: str, value: Any) -> Dict:
        """
        Update an attribute of a given node.

        Parameters
        ----------
        node: str
            The node identifier
        key: str
            The key for an attribute
        value: Any
            The value corresponding to the key

        Returns
        -------
        Dict
            A dictionary corresponding to the updated node properties

        """
        pass

    def update_edge_attribute(
        self,
        subject_node: str,
        object_node: str,
        edge_key: Optional[str],
        attr_key: str,
        attr_value: Any,
    ) -> Dict:
        """
        Update an attribute of a given edge.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key
        attr_key: str
            The attribute key
        attr_value: Any
            The attribute value

        Returns
        -------
        Dict
            A dictionary corresponding to the updated edge properties

        """
        pass

    def get_node(self, node: str) -> Dict:
        """
        Get a node and its properties.

        Parameters
        ----------
        node: str
            The node identifier

        Returns
        -------
        Dict
            The node dictionary

        """
        pass

    def get_edge(
        self, subject_node: str, object_node: str, edge_key: Optional[str]
    ) -> Dict:
        """
        Get an edge and its properties.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key

        Returns
        -------
        Dict
            The edge dictionary

        """
        pass

    def nodes(self, data: bool = True) -> Dict:
        """
        Get all nodes in a graph.

        Parameters
        ----------
        data: bool
            Whether or not to fetch node properties

        Returns
        -------
        Dict
            A dictionary of nodes

        """
        pass

    def edges(self, keys: bool = False, data: bool = True) -> Dict:
        """
        Get all edges in a graph.

        Parameters
        ----------
        keys: bool
            Whether or not to include edge keys
        data: bool
            Whether or not to fetch node properties

        Returns
        -------
        Dict
            A dictionary of edges

        """
        pass

    def in_edges(self, node: str, keys: bool = False, data: bool = False) -> List:
        """
        Get all incoming edges for a given node.

        Parameters
        ----------
        node: str
            The node identifier
        keys: bool
            Whether or not to include edge keys
        data: bool
            Whether or not to fetch node properties

        Returns
        -------
        List
            A list of edges

        """
        pass

    def out_edges(self, node: str, keys: bool = False, data: bool = False) -> List:
        """
        Get all outgoing edges for a given node.

        Parameters
        ----------
        node: str
            The node identifier
        keys: bool
            Whether or not to include edge keys
        data: bool
            Whether or not to fetch node properties

        Returns
        -------
        List
            A list of edges

        """
        pass

    def nodes_iter(self) -> Generator:
        """
        Get an iterable to traverse through all the nodes in a graph.

        Returns
        -------
        Generator
            A generator for nodes

        """
        pass

    def edges_iter(self) -> Generator:
        """
        Get an iterable to traverse through all the edges in a graph.

        Returns
        -------
        Generator
            A generator for edges

        """
        for u, v, k, data in self.edges(keys=True, data=True):
            yield (u, v, k, data)

    def remove_node(self, node: str) -> Any:
        """
        Remove a given node from the graph.

        Parameters
        ----------
        node: str
            The node identifier

        Returns
        -------
        Any

        """
        pass

    def remove_edge(
        self, subject_node: str, object_node: str, edge_key: Optional[str] = None
    ) -> Any:
        """
        Remove a given edge from the graph.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key

        Returns
        -------
        Any

        """
        pass

    def has_node(self, node: str) -> bool:
        """
        Check whether a given node exists in the graph.

        Parameters
        ----------
        node: str
            The node identifier

        Returns
        -------
        bool
            Whether or not the given node exists

        """
        pass

    def has_edge(
        self, subject_node: str, object_node: str, edge_key: Optional[str] = None
    ) -> bool:
        """
        Check whether a given edge exists in the graph.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key

        Returns
        -------
        bool
            Whether or not the given edge exists

        """
        pass

    def number_of_nodes(self) -> int:
        """
        Returns the number of nodes in a graph.

        Returns
        -------
        int

        """
        pass

    def number_of_edges(self) -> int:
        """
        Returns the number of edges in a graph.

        Returns
        -------
        int

        """
        pass

    def degree(self):
        """
        Get the degree of all the nodes in a graph.
        """
        pass

    def clear(self) -> None:
        """
        Remove all the nodes and edges in the graph.
        """
        pass

    @staticmethod
    def set_node_attributes(graph: Any, attributes: Dict) -> Any:
        """
        Set nodes attributes from a dictionary of key-values.

        Parameters
        ----------
        graph: Any
            The graph to modify
        attributes: Dict
            A dictionary of node identifier to key-value pairs

        Returns
        -------
        Any

        """
        pass

    @staticmethod
    def set_edge_attributes(graph: Any, attributes: Dict) -> Any:
        """
        Set nodes attributes from a dictionary of key-values.

        Parameters
        ----------
        graph: Any
            The graph to modify
        attributes: Dict
            A dictionary of node identifier to key-value pairs

        Returns
        -------
        Any

        """
        pass

    @staticmethod
    def get_node_attributes(graph: Any, attr_key: str) -> Any:
        """
        Get all nodes that have a value for the given attribute ``attr_key``.

        Parameters
        ----------
        graph: Any
            The graph to modify
        attr_key: str
            The attribute key

        Returns
        -------
        Any

        """
        pass

    @staticmethod
    def get_edge_attributes(graph: Any, attr_key: str) -> Any:
        """
        Get all edges that have a value for the given attribute ``attr_key``.

        Parameters
        ----------
        graph: Any
            The graph to modify
        attr_key: str
            The attribute key

        Returns
        -------
        Any

        """
        pass

    @staticmethod
    def relabel_nodes(graph: Any, mapping: Dict) -> Any:
        """
        Relabel identifiers for a series of nodes based on mappings.

        Parameters
        ----------
        graph: Any
            The graph to modify
        mapping: Dict[str, str]
            A dictionary of mapping where the key is the old identifier
            and the value is the new identifier.

        Returns
        -------
        Any

        """
        pass



import re
from enum import Enum
from typing import List, TextIO, Optional, Dict, Set, Callable

import click
import validators
from bmt import Toolkit

from kgx.config import get_jsonld_context, get_logger
from kgx.graph.base_graph import BaseGraph
from kgx.utils.kgx_utils import (
    get_toolkit,
    snakecase_to_sentencecase,
    sentencecase_to_snakecase,
    camelcase_to_sentencecase,
    GraphEntityType,
)
from kgx.prefix_manager import PrefixManager

logger = get_logger()


class ErrorType(Enum):
    """
    Validation error types
    """

    MISSING_NODE_PROPERTY = 1
    MISSING_EDGE_PROPERTY = 2
    INVALID_NODE_PROPERTY_VALUE_TYPE = 3
    INVALID_NODE_PROPERTY_VALUE = 4
    INVALID_EDGE_PROPERTY_VALUE_TYPE = 5
    INVALID_EDGE_PROPERTY_VALUE = 6
    NO_CATEGORY = 7
    INVALID_CATEGORY = 8
    NO_EDGE_PREDICATE = 9
    INVALID_EDGE_PREDICATE = 10


class MessageLevel(Enum):
    """
    Message level for validation reports
    """

    # Recommendations
    INFO = 1
    # Message to convey 'should'
    WARNING = 2
    # Message to convey 'must'
    ERROR = 3


class ValidationError(object):
    """
    ValidationError class that represents an error.

    Parameters
    ----------
    entity: str
        The node or edge entity that is failing validation
    error_type: kgx.validator.ErrorType
        The nature of the error
    message: str
        The error message
    message_level: kgx.validator.MessageLevel
        The message level

    """

    def __init__(
        self,
        entity: str,
        error_type: ErrorType,
        message: str,
        message_level: MessageLevel,
    ):
        self.entity = entity
        self.error_type = error_type
        self.message = message
        self.message_level = message_level

    def __str__(self):
        return f"[{self.message_level.name}][{self.error_type.name}] {self.entity} - {self.message}"

    def as_dict(self):
        return {
            "entity": self.entity,
            "error_type": self.error_type.name,
            "message": self.message,
            "message_level": self.message_level.name,
        }


class Validator(object):
    """
    Class for validating a property graph.

    The optional 'progress_monitor' for the validator should be a lightweight Callable
    which is injected into the class 'inspector' Callable, designed to intercepts
    node and edge records streaming through the Validator (inside a Transformer.process() call.
    The first (GraphEntityType) argument of the Callable tags the record as a NODE or an EDGE.
    The second argument given to the Callable is the current record itself.
    This Callable is strictly meant to be procedural and should *not* mutate the record.
    The intent of this Callable is to provide a hook to KGX applications wanting the
    namesake function of passively monitoring the graph data stream. As such, the Callable
    could simply tally up the number of times it is called with a NODE or an EDGE, then
    provide a suitable (quick!) report of that count back to the KGX application. The
    Callable (function/callable class) should not modify the record and should be of low
    complexity, so as not to introduce a large computational overhead to validation!

    Parameters
    ----------
    verbose: bool
        Whether the generated report should be verbose or not (default: ``False``)
    progress_monitor: Optional[Callable[[GraphEntityType, List], None]]
        Function given a peek at the current record being processed by the class wrapped Callable.
    schema: Optional[str]
        URL to (Biolink) Model Schema to be used for validated (default: None, use default Biolink Model Toolkit schema)
    """

    def __init__(
        self,
        verbose: bool = False,
        progress_monitor: Optional[Callable[[GraphEntityType, List], None]] = None,
        schema: Optional[str] = None,
    ):
        # formal arguments
        self.verbose: bool = verbose
        self.progress_monitor: Optional[
            Callable[[GraphEntityType, List], None]
        ] = progress_monitor

        # internal attributes
        # associated currently active _currently_active_toolkit with this Validator instance
        self.validating_toolkit = self.get_toolkit()
        self.prefix_manager = PrefixManager()
        self.jsonld = get_jsonld_context()
        self.prefixes = Validator.get_all_prefixes(self.jsonld)
        self.required_node_properties = Validator.get_required_node_properties()
        self.required_edge_properties = Validator.get_required_edge_properties()
        self.errors: List[ValidationError] = list()

    def __call__(self, entity_type: GraphEntityType, rec: List):
        """
        Transformer 'inspector' Callable
        """
        if self.progress_monitor:
            self.progress_monitor(entity_type, rec)
        if entity_type == GraphEntityType.EDGE:
            self.errors += self.analyse_edge(*rec)
        elif entity_type == GraphEntityType.NODE:
            self.errors += self.analyse_node(*rec)
        else:
            raise RuntimeError("Unexpected GraphEntityType: " + str(entity_type))

    def get_validating_toolkit(self):
        return self.validating_toolkit

    def get_validation_model_version(self):
        return self.validating_toolkit.get_model_version()

    def get_errors(self):
        return self.errors

    _currently_active_toolkit: Optional[Toolkit] = None

    @classmethod
    def set_biolink_model(cls, version: Optional[str]):
        cls._currently_active_toolkit = get_toolkit(biolink_release=version)

    @classmethod
    def get_toolkit(cls) -> Toolkit:
        if not cls._currently_active_toolkit:
            cls._currently_active_toolkit = get_toolkit()
        return cls._currently_active_toolkit

    _default_model_version = None

    @classmethod
    def get_default_model_version(cls):
        if not cls._default_model_version:
            # get default Biolink version from BMT
            cls._default_model_version = get_toolkit().get_model_version()
        return cls._default_model_version

    def analyse_node(self, n, data):
        e1 = Validator.validate_node_properties(n, data, self.required_node_properties)
        e2 = Validator.validate_node_property_types(
            n, data, toolkit=self.validating_toolkit
        )
        e3 = Validator.validate_node_property_values(n, data)
        e4 = Validator.validate_categories(n, data, toolkit=self.validating_toolkit)
        return e1 + e2 + e3 + e4

    def analyse_edge(self, u, v, k, data):
        e1 = Validator.validate_edge_properties(
            u, v, data, self.required_edge_properties
        )
        e2 = Validator.validate_edge_property_types(
            u, v, data, toolkit=self.validating_toolkit
        )
        e3 = Validator.validate_edge_property_values(u, v, data)
        e4 = Validator.validate_edge_predicate(
            u, v, data, toolkit=self.validating_toolkit
        )
        return e1 + e2 + e3 + e4

    @staticmethod
    def get_all_prefixes(jsonld: Optional[Dict] = None) -> set:
        """
        Get all prefixes from Biolink Model JSON-LD context.

        It also sets ``self.prefixes`` for subsequent access.

        Parameters
        ---------
        jsonld: Optional[Dict]
            The JSON-LD context

        Returns
        -------
        Optional[Dict]
            A set of prefixes

        """
        if not jsonld:
            jsonld = get_jsonld_context()
        prefixes: Set = set(
            k
            for k, v in jsonld.items()
            if isinstance(v, str)
            or (isinstance(v, dict) and v.setdefault("@prefix", False))
        )  # @type: ignored
        if "biolink" not in prefixes:
            prefixes.add("biolink")
        return prefixes

    @staticmethod
    def get_required_node_properties(toolkit: Optional[Toolkit] = None) -> list:
        """
        Get all properties for a node that are required, as defined by Biolink Model.

        Parameters
        ----------
        toolkit: Optional[Toolkit]
            Optional externally provided toolkit (default: use Validator class defined toolkit)

        Returns
        -------
        list
            A list of required node properties

        """
        if not toolkit:
            toolkit = Validator.get_toolkit()
        node_properties = toolkit.get_all_node_properties()
        required_properties = []
        for p in node_properties:
            element = toolkit.get_element(p)
            if element and element.deprecated is None:
                if hasattr(element, "required") and element.required:
                    formatted_name = sentencecase_to_snakecase(element.name)
                    required_properties.append(formatted_name)
                elif element.name == "category":
                    formatted_name = sentencecase_to_snakecase(element.name)
                    required_properties.append(formatted_name)
        return required_properties

    @staticmethod
    def get_required_edge_properties(toolkit: Optional[Toolkit] = None) -> list:
        """
        Get all properties for an edge that are required, as defined by Biolink Model.

        Parameters
        ----------
        toolkit: Optional[Toolkit]
            Optional externally provided toolkit (default: use Validator class defined toolkit)

        Returns
        -------
        list
            A list of required edge properties

        """
        if not toolkit:
            toolkit = Validator.get_toolkit()
        edge_properties = toolkit.get_all_edge_properties()
        required_properties = []
        for p in edge_properties:
            element = toolkit.get_element(p)
            if element and element.deprecated is None:
                if hasattr(element, "required") and element.required:
                    formatted_name = sentencecase_to_snakecase(element.name)
                    required_properties.append(formatted_name)
        return required_properties

    def validate(self, graph: BaseGraph) -> list:
        """
        Validate nodes and edges in a graph.
        TODO: Support strict mode

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to validate

        Returns
        -------
        list
            A list of errors for a given graph

        """
        node_errors = self.validate_nodes(graph)
        edge_errors = self.validate_edges(graph)
        self.errors = node_errors + edge_errors
        return self.errors

    def validate_nodes(self, graph: BaseGraph) -> list:
        """
        Validate all the nodes in a graph.

        This method validates for the following,
        - Node properties
        - Node property type
        - Node property value type
        - Node categories

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to validate

        Returns
        -------
        list
            A list of errors for a given graph

        """
        errors = []
        with click.progressbar(
            graph.nodes(data=True), label="Validating nodes in graph"
        ) as bar:
            for n, data in bar:
                errors += self.analyse_node(n, data)
        return errors

    def validate_edges(self, graph: BaseGraph) -> list:
        """
        Validate all the edges in a graph.

        This method validates for the following,
        - Edge properties
        - Edge property type
        - Edge property value type
        - Edge predicate

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to validate

        Returns
        -------
        list
            A list of errors for a given graph

        """
        errors = []
        with click.progressbar(
            graph.edges(data=True), label="Validate edges in graph"
        ) as bar:
            for u, v, data in bar:
                errors += self.analyse_edge(u, v, None, data)
        return errors

    @staticmethod
    def validate_node_properties(
        node: str, data: dict, required_properties: list
    ) -> list:
        """
        Checks if all the required node properties exist for a given node.

        Parameters
        ----------
        node: str
            Node identifier
        data: dict
            Node properties
        required_properties: list
            Required node properties

        Returns
        -------
        list
            A list of errors for a given node

        """
        errors = []
        for p in required_properties:
            if p not in data:
                error_type = ErrorType.MISSING_NODE_PROPERTY
                message = f"Required node property '{p}' missing"
                errors.append(
                    ValidationError(node, error_type, message, MessageLevel.ERROR)
                )
        return errors

    @staticmethod
    def validate_edge_properties(
        subject: str, object: str, data: dict, required_properties: list
    ) -> list:
        """
        Checks if all the required edge properties exist for a given edge.

        Parameters
        ----------
        subject: str
            Subject identifier
        object: str
            Object identifier
        data: dict
            Edge properties
        required_properties: list
            Required edge properties

        Returns
        -------
        list
            A list of errors for a given edge

        """
        errors = []
        for p in required_properties:
            if p not in data:
                if p == "association_id":
                    # check for 'id' property instead
                    if "id" not in data:
                        error_type = ErrorType.MISSING_EDGE_PROPERTY
                        message = f"Required edge property '{p}' missing"
                        errors.append(
                            ValidationError(
                                f"{subject}-{object}",
                                error_type,
                                message,
                                MessageLevel.ERROR,
                            )
                        )
                else:
                    error_type = ErrorType.MISSING_EDGE_PROPERTY
                    message = f"Required edge property '{p}' missing"
                    errors.append(
                        ValidationError(
                            f"{subject}-{object}",
                            error_type,
                            message,
                            MessageLevel.ERROR,
                        )
                    )
        return errors

    @staticmethod
    def validate_node_property_types(
        node: str, data: dict, toolkit: Optional[Toolkit] = None
    ) -> list:
        """
        Checks if node properties have the expected value type.

        Parameters
        ----------
        node: str
            Node identifier
        data: dict
            Node properties
        toolkit: Optional[Toolkit]
            Optional externally provided toolkit (default: use Validator class defined toolkit)

        Returns
        -------
        list
            A list of errors for a given node

        """
        if not toolkit:
            toolkit = Validator.get_toolkit()
        errors = []
        error_type = ErrorType.INVALID_NODE_PROPERTY_VALUE_TYPE
        if not isinstance(node, str):
            message = "Node property 'id' expected to be of type 'string'"
            errors.append(
                ValidationError(node, error_type, message, MessageLevel.ERROR)
            )

        for key, value in data.items():
            element = toolkit.get_element(key)
            if element:
                if hasattr(element, "typeof"):
                    if element.typeof == "string" and not isinstance(value, str):
                        message = f"Node property '{key}' expected to be of type '{element.typeof}'"
                        errors.append(
                            ValidationError(
                                node, error_type, message, MessageLevel.ERROR
                            )
                        )
                    elif (
                        element.typeof == "uriorcurie"
                        and not isinstance(value, str)
                        and not validators.url(value)
                    ):
                        message = f"Node property '{key}' expected to be of type 'uri' or 'CURIE'"
                        errors.append(
                            ValidationError(
                                node, error_type, message, MessageLevel.ERROR
                            )
                        )
                    elif element.typeof == "double" and not isinstance(
                        value, (int, float)
                    ):
                        message = f"Node property '{key}' expected to be of type '{element.typeof}'"
                        errors.append(
                            ValidationError(
                                node, error_type, message, MessageLevel.ERROR
                            )
                        )
                    else:
                        logger.warning(
                            "Skipping validation for Node property '{}'. Expected type '{}' vs Actual type '{}'".format(
                                key, element.typeof, type(value)
                            )
                        )
                if hasattr(element, "multivalued"):
                    if element.multivalued:
                        if not isinstance(value, list):
                            message = f"Multi-valued node property '{key}' expected to be of type '{list}'"
                            errors.append(
                                ValidationError(
                                    node, error_type, message, MessageLevel.ERROR
                                )
                            )
                    else:
                        if isinstance(value, (list, set, tuple)):
                            message = f"Single-valued node property '{key}' expected to be of type '{str}'"
                            errors.append(
                                ValidationError(
                                    node, error_type, message, MessageLevel.ERROR
                                )
                            )
        return errors

    @staticmethod
    def validate_edge_property_types(
        subject: str, object: str, data: dict, toolkit: Optional[Toolkit] = None
    ) -> list:
        """
        Checks if edge properties have the expected value type.

        Parameters
        ----------
        subject: str
            Subject identifier
        object: str
            Object identifier
        data: dict
            Edge properties
        toolkit: Optional[Toolkit]
            Optional externally provided toolkit (default: use Validator class defined toolkit)

        Returns
        -------
        list
            A list of errors for a given edge

        """
        if not toolkit:
            toolkit = Validator.get_toolkit()
        errors = []
        error_type = ErrorType.INVALID_EDGE_PROPERTY_VALUE_TYPE
        if not isinstance(subject, str):
            message = "'subject' of an edge expected to be of type 'string'"
            errors.append(
                ValidationError(
                    f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                )
            )
        if not isinstance(object, str):
            message = "'object' of an edge expected to be of type 'string'"
            errors.append(
                ValidationError(
                    f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                )
            )

        for key, value in data.items():
            element = toolkit.get_element(key)
            if element:
                if hasattr(element, "typeof"):
                    if element.typeof == "string" and not isinstance(value, str):
                        message = (
                            f"Edge property '{key}' expected to be of type 'string'"
                        )
                        errors.append(
                            ValidationError(
                                f"{subject}-{object}",
                                error_type,
                                message,
                                MessageLevel.ERROR,
                            )
                        )
                    elif (
                        element.typeof == "uriorcurie"
                        and not isinstance(value, str)
                        and not validators.url(value)
                    ):
                        message = f"Edge property '{key}' expected to be of type 'uri' or 'CURIE'"
                        errors.append(
                            ValidationError(
                                f"{subject}-{object}",
                                error_type,
                                message,
                                MessageLevel.ERROR,
                            )
                        )
                    elif element.typeof == "double" and not isinstance(
                        value, (int, float)
                    ):
                        message = (
                            f"Edge property '{key}' expected to be of type 'double'"
                        )
                        errors.append(
                            ValidationError(
                                f"{subject}-{object}",
                                error_type,
                                message,
                                MessageLevel.ERROR,
                            )
                        )
                    else:
                        logger.warning(
                            "Skipping validation for Edge property '{}'. Expected type '{}' vs Actual type '{}'".format(
                                key, element.typeof, type(value)
                            )
                        )
                if hasattr(element, "multivalued"):
                    if element.multivalued:
                        if not isinstance(value, list):
                            message = f"Multi-valued edge property '{key}' expected to be of type 'list'"
                            errors.append(
                                ValidationError(
                                    f"{subject}-{object}",
                                    error_type,
                                    message,
                                    MessageLevel.ERROR,
                                )
                            )
                    else:
                        if isinstance(value, (list, set, tuple)):
                            message = f"Single-valued edge property '{key}' expected to be of type 'str'"
                            errors.append(
                                ValidationError(
                                    f"{subject}-{object}",
                                    error_type,
                                    message,
                                    MessageLevel.ERROR,
                                )
                            )
        return errors

    @staticmethod
    def validate_node_property_values(node: str, data: dict) -> list:
        """
        Validate a node property's value.

        Parameters
        ----------
        node: str
            Node identifier
        data: dict
            Node properties

        Returns
        -------
        list
            A list of errors for a given node

        """
        errors = []
        error_type = ErrorType.INVALID_NODE_PROPERTY_VALUE
        if not PrefixManager.is_curie(node):
            message = f"Node property 'id' expected to be of type 'CURIE'"
            errors.append(
                ValidationError(node, error_type, message, MessageLevel.ERROR)
            )
        else:
            prefix = PrefixManager.get_prefix(node)
            if prefix and prefix not in Validator.get_all_prefixes():
                message = f"Node property 'id' has a value '{node}' with a CURIE prefix '{prefix}' is not represented in Biolink Model JSON-LD context"
                errors.append(
                    ValidationError(node, error_type, message, MessageLevel.ERROR)
                )
        return errors

    @staticmethod
    def validate_edge_property_values(subject: str, object: str, data: dict) -> list:
        """
        Validate an edge property's value.

        Parameters
        ----------
        subject: str
            Subject identifier
        object: str
            Object identifier
        data: dict
            Edge properties

        Returns
        -------
        list
            A list of errors for a given edge

        """
        errors = []
        error_type = ErrorType.INVALID_EDGE_PROPERTY_VALUE
        prefixes = Validator.get_all_prefixes()

        if PrefixManager.is_curie(subject):
            prefix = PrefixManager.get_prefix(subject)
            if prefix and prefix not in prefixes:
                message = f"Edge property 'subject' has a value '{subject}' with a CURIE prefix '{prefix}' that is not represented in Biolink Model JSON-LD context"
                errors.append(
                    ValidationError(
                        f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                    )
                )
        else:
            message = f"Edge property 'subject' has a value '{subject}' which is not a proper CURIE"
            errors.append(
                ValidationError(
                    f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                )
            )

        if PrefixManager.is_curie(object):
            prefix = PrefixManager.get_prefix(object)
            if prefix not in prefixes:
                message = f"Edge property 'object' has a value '{object}' with a CURIE prefix '{prefix}' that is not represented in Biolink Model JSON-LD context"
                errors.append(
                    ValidationError(
                        f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                    )
                )
        else:
            message = f"Edge property 'object' has a value '{object}' which is not a proper CURIE"
            errors.append(
                ValidationError(
                    f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                )
            )
        if "relation" in data:
            if PrefixManager.is_curie(data["relation"]):
                prefix = PrefixManager.get_prefix(data["relation"])
                if prefix not in prefixes:
                    message = f"Edge property 'relation' has a value '{data['relation']}' with a CURIE prefix '{prefix}' that is not represented in Biolink Model JSON-LD context"
                    errors.append(
                        ValidationError(
                            f"{subject}-{object}",
                            error_type,
                            message,
                            MessageLevel.ERROR,
                        )
                    )
            else:
                message = f"Edge property 'relation' has a value '{data['relation']}' which is not a proper CURIE"
                errors.append(
                    ValidationError(
                        f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                    )
                )
        return errors

    @staticmethod
    def validate_categories(
        node: str, data: dict, toolkit: Optional[Toolkit] = None
    ) -> list:
        """
        Validate ``category`` field of a given node.

        Parameters
        ----------
        node: str
            Node identifier
        data: dict
            Node properties
        toolkit: Optional[Toolkit]
            Optional externally provided toolkit (default: use Validator class defined toolkit)

        Returns
        -------
        list
            A list of errors for a given node

        """
        if not toolkit:
            toolkit = Validator.get_toolkit()
        error_type = ErrorType.INVALID_CATEGORY
        errors = []
        categories = data.get("category")
        if categories is None:
            message = "Node does not have a 'category' property"
            errors.append(
                ValidationError(node, error_type, message, MessageLevel.ERROR)
            )
        elif not isinstance(categories, list):
            message = f"Node property 'category' expected to be of type {list}"
            errors.append(
                ValidationError(node, error_type, message, MessageLevel.ERROR)
            )
        else:
            for category in categories:
                if PrefixManager.is_curie(category):
                    category = PrefixManager.get_reference(category)
                m = re.match(r"^([A-Z][a-z\d]+)+$", category)
                if not m:
                    # category is not CamelCase
                    error_type = ErrorType.INVALID_CATEGORY
                    message = f"Category '{category}' is not in CamelCase form"
                    errors.append(
                        ValidationError(node, error_type, message, MessageLevel.ERROR)
                    )
                formatted_category = camelcase_to_sentencecase(category)
                if toolkit.is_mixin(formatted_category):
                    message = f"Category '{category}' is a mixin in the Biolink Model"
                    errors.append(
                        ValidationError(node, error_type, message, MessageLevel.ERROR)
                    )
                elif not toolkit.is_category(formatted_category):
                    message = (
                        f"Category '{category}' unknown in the current Biolink Model"
                    )
                    errors.append(
                        ValidationError(node, error_type, message, MessageLevel.ERROR)
                    )
                else:
                    c = toolkit.get_element(formatted_category.lower())
                    if c:
                        if category != c.name and category in c.aliases:
                            message = f"Category {category} is actually an alias for {c.name}; Should replace '{category}' with '{c.name}'"
                            errors.append(
                                ValidationError(
                                    node, error_type, message, MessageLevel.ERROR
                                )
                            )
        return errors

    @staticmethod
    def validate_edge_predicate(
        subject: str, object: str, data: dict, toolkit: Optional[Toolkit] = None
    ) -> list:
        """
        Validate ``edge_predicate`` field of a given edge.

        Parameters
        ----------
        subject: str
            Subject identifier
        object: str
            Object identifier
        data: dict
            Edge properties
        toolkit: Optional[Toolkit]
            Optional externally provided toolkit (default: use Validator class defined toolkit)

        Returns
        -------
        list
            A list of errors for a given edge

        """
        if not toolkit:
            toolkit = Validator.get_toolkit()
        error_type = ErrorType.INVALID_EDGE_PREDICATE
        errors = []
        edge_predicate = data.get("predicate")
        if edge_predicate is None:
            message = "Edge does not have an 'predicate' property"
            errors.append(
                ValidationError(
                    f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                )
            )
        elif not isinstance(edge_predicate, str):
            message = f"Edge property 'edge_predicate' expected to be of type 'string'"
            errors.append(
                ValidationError(
                    f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                )
            )
        else:
            if PrefixManager.is_curie(edge_predicate):
                edge_predicate = PrefixManager.get_reference(edge_predicate)
            m = re.match(r"^([a-z_][^A-Z\s]+_?[a-z_][^A-Z\s]+)+$", edge_predicate)
            if m:
                p = toolkit.get_element(snakecase_to_sentencecase(edge_predicate))
                if p is None:
                    message = f"Edge predicate '{edge_predicate}' not in Biolink Model"
                    errors.append(
                        ValidationError(
                            f"{subject}-{object}",
                            error_type,
                            message,
                            MessageLevel.ERROR,
                        )
                    )
                elif edge_predicate != p.name and edge_predicate in p.aliases:
                    message = f"Edge predicate '{edge_predicate}' is actually an alias for {p.name}; Should replace {edge_predicate} with {p.name}"
                    errors.append(
                        ValidationError(
                            f"{subject}-{object}",
                            error_type,
                            message,
                            MessageLevel.ERROR,
                        )
                    )
            else:
                message = f"Edge predicate '{edge_predicate}' is not in snake_case form"
                errors.append(
                    ValidationError(
                        f"{subject}-{object}", error_type, message, MessageLevel.ERROR
                    )
                )
        return errors

    @staticmethod
    def report(errors: List[ValidationError]) -> List:
        """
        Prepare error report.

        Parameters
        ----------
        errors: List[ValidationError]
            List of kgx.validator.ValidationError

        Returns
        -------
        List
            A list of formatted errors

        """
        return [str(x) for x in errors]

    def get_error_messages(self):
        """
        A direct Validator "instance" method version of report()
        that directly accesses the internal Validator self.errors list.

        Returns
        -------
        List
            A list of formatted error messages.

        """
        return Validator.report(self.errors)

    def write_report(self, outstream: TextIO) -> None:
        """
        Write error report to a file

        Parameters
        ----------
        outstream: TextIO
            The stream to write to

        """
        for x in Validator.report(self.errors):
            outstream.write(f"{x}\n")



import re
from typing import Dict, Optional, Any

import prefixcommons.curie_util as cu
from cachetools import LRUCache, cached

from kgx.config import get_jsonld_context, get_logger
from kgx.utils.kgx_utils import contract, expand

log = get_logger()


class PrefixManager(object):
    """
    Manages prefix mappings.

    These include mappings for CURIEs such as GO:0008150, as well as shortforms such as
    biolink types such as Disease
    """

    DEFAULT_NAMESPACE = "https://www.example.org/UNKNOWN/"
    prefix_map: Dict[str, str]
    reverse_prefix_map: Dict[str, str]

    def __init__(self, url: str = None):
        """
        Initialize an instance of PrefixManager.

        Parameters
        ----------
        url: str
            The URL from which to read a JSON-LD context for prefix mappings

        """
        if url:
            context = cu.read_remote_jsonld_context(url)
        else:
            context = get_jsonld_context()
        self.set_prefix_map(context)

    def set_prefix_map(self, m: Dict) -> None:
        """
        Populate `prefix_map` with contents from a JSON-LD context from self.url

        Parameters
        ----------
        m: dict
            Dictionary of prefix to URI mappings

        """
        self.prefix_map = {}
        for k, v in m.items():
            if isinstance(v, str):
                self.prefix_map[k] = v
            else:
                self.prefix_map[k] = v.get("@id")

        if "biolink" not in self.prefix_map:
            self.prefix_map["biolink"] = (
                self.prefix_map["@vocab"]
                if "@vocab" in self.prefix_map
                else "https://w3id.org/biolink/vocab/"
            )
        if "owlstar" not in self.prefix_map:
            self.prefix_map["owlstar"] = "http://w3id.org/owlstar/"
        if "@vocab" in self.prefix_map:
            del self.prefix_map["@vocab"]
        if "MONARCH" not in self.prefix_map:
            self.prefix_map["MONARCH"] = "https://monarchinitiative.org/"
            self.prefix_map["MONARCH_NODE"] = "https://monarchinitiative.org/MONARCH_"
        if "" in self.prefix_map:
            log.info(
                f"Replacing default prefix mapping from {self.prefix_map['']} to 'www.example.org/UNKNOWN/'"
            )
        else:
            self.prefix_map[""] = self.DEFAULT_NAMESPACE
        self.reverse_prefix_map = {y: x for x, y in self.prefix_map.items()}

    def update_prefix_map(self, m: Dict[str, str]) -> None:
        """
        Update prefix maps with new mappings.

        Parameters
        ----------
        m: Dict
            New prefix to IRI mappings

        """
        for k, v in m.items():
            self.prefix_map[k] = v

    def update_reverse_prefix_map(self, m: Dict[str, str]) -> None:
        """
        Update reverse prefix maps with new mappings.

        Parameters
        ----------
        m: Dict
            New IRI to prefix mappings

        """
        self.reverse_prefix_map.update(m)

    @cached(LRUCache(maxsize=1024))
    def expand(self, curie: str, fallback: bool = True) -> str:
        """
        Expand a given CURIE to an URI, based on mappings from `prefix_map`.

        Parameters
        ----------
        curie: str
            A CURIE
        fallback: bool
            Determines whether to fallback to default prefix mappings, as determined
            by `prefixcommons.curie_util`, when CURIE prefix is not found in `prefix_map`.

        Returns
        -------
        str
            A URI corresponding to the CURIE

        """
        uri = expand(curie, [self.prefix_map], fallback)
        return uri

    @cached(LRUCache(maxsize=1024))
    def contract(self, uri: str, fallback: bool = True) -> Optional[str]:
        """
        Contract a given URI to a CURIE, based on mappings from `prefix_map`.

        Parameters
        ----------
        uri: str
            A URI

        fallback: bool
            Determines whether to fallback to default prefix mappings, as determined
            by `prefixcommons.curie_util`, when URI prefix is not found in `reverse_prefix_map`.

        Returns
        -------
        Optional[str]
            A CURIE corresponding to the URI

        """
        # always prioritize non-CURIE shortform
        if self.reverse_prefix_map and uri in self.reverse_prefix_map:
            curie = self.reverse_prefix_map[uri]
        else:
            curie = contract(uri, [self.prefix_map], fallback)
        return str(curie)

    @staticmethod
    @cached(LRUCache(maxsize=1024))
    def is_curie(s: str) -> bool:
        """
        Check if a given string is a CURIE.

        Parameters
        ----------
        s: str
            A string

        Returns
        -------
        bool
            Whether or not the given string is a CURIE

        """
        if isinstance(s, str):
            m = re.match(r"^[^ <()>:]*:[^/ :]+$", s)
            return bool(m)
        else:
            return False

    @staticmethod
    @cached(LRUCache(maxsize=1024))
    def is_iri(s: str) -> bool:
        """
        Check if a given string as an IRI.

        Parameters
        ----------
        s: str
            A string

        Returns
        -------
        bool
            Whether or not the given string is an IRI.

        """
        if isinstance(s, str):
            return s.startswith("http") or s.startswith("https")
        else:
            return False

    @staticmethod
    @cached(LRUCache(maxsize=1024))
    def has_urlfragment(s: str) -> bool:
        if "#" in s:
            return True
        else:
            return False

    @staticmethod
    @cached(LRUCache(maxsize=1024))
    def get_prefix(curie: str) -> Optional[str]:
        """
        Get the prefix from a given CURIE.

        Parameters
        ----------
        curie: str
            The CURIE

        Returns
        -------
        str
            The CURIE prefix

        """
        prefix: Optional[str] = None
        if PrefixManager.is_curie(curie):
            prefix = curie.split(":", 1)[0]
        return prefix

    @staticmethod
    @cached(LRUCache(maxsize=1024))
    def get_reference(curie: str) -> Optional[str]:
        """
        Get the reference of a given CURIE.

        Parameters
        ----------
        curie: str
            The CURIE

        Returns
        -------
        Optional[str]
            The reference of a CURIE

        """
        reference: Optional[str] = None
        if PrefixManager.is_curie(curie):
            reference = curie.split(":", 1)[1]
        return reference



from .sink import Sink
from .tsv_sink import TsvSink
from .json_sink import JsonSink
from .jsonl_sink import JsonlSink
from .neo_sink import NeoSink
from .rdf_sink import RdfSink
from .graph_sink import GraphSink
from .null_sink import NullSink



import copy
from typing import Tuple, Optional, Dict, List, Any, Set, Union

import networkx as nx
from ordered_set import OrderedSet

from kgx.config import get_logger
from kgx.graph.base_graph import BaseGraph
from kgx.utils.kgx_utils import (
    get_prefix_prioritization_map,
    get_biolink_element,
    get_biolink_ancestors,
    current_time_in_millis,
    format_biolink_category,
    generate_edge_key,
    get_toolkit,
)

log = get_logger()
toolkit = get_toolkit()
SAME_AS = "biolink:same_as"
SUBCLASS_OF = "biolink:subclass_of"
LEADER_ANNOTATION = "clique_leader"
ORIGINAL_SUBJECT_PROPERTY = "_original_subject"
ORIGINAL_OBJECT_PROPERTY = "_original_object"


def clique_merge(
    target_graph: BaseGraph,
    leader_annotation: str = None,
    prefix_prioritization_map: Optional[Dict[str, List[str]]] = None,
    category_mapping: Optional[Dict[str, str]] = None,
    strict: bool = True,
) -> Tuple[BaseGraph, nx.MultiDiGraph]:
    """

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        The original graph
    leader_annotation: str
        The field on a node that signifies that the node is the leader of a clique
    prefix_prioritization_map: Optional[Dict[str, List[str]]]
        A map that gives a prefix priority for one or more categories
    category_mapping: Optional[Dict[str, str]]
        Mapping for non-Biolink Model categories to Biolink Model categories
    strict: bool
        Whether or not to merge nodes in a clique that have conflicting node categories

    Returns
    -------
    Tuple[kgx.graph.base_graph.BaseGraph, networkx.MultiDiGraph]
        A tuple containing the updated target graph, and the clique graph

    """
    ppm = get_prefix_prioritization_map()
    if prefix_prioritization_map:
        ppm.update(prefix_prioritization_map)
    prefix_prioritization_map = ppm

    if not leader_annotation:
        leader_annotation = LEADER_ANNOTATION

    start = current_time_in_millis()
    clique_graph = build_cliques(target_graph)
    end = current_time_in_millis()
    log.info(f"Total time taken to build cliques: {end - start} ms")

    start = current_time_in_millis()
    elect_leader(
        target_graph,
        clique_graph,
        leader_annotation,
        prefix_prioritization_map,
        category_mapping,
        strict,
    )
    end = current_time_in_millis()
    log.info(f"Total time taken to elect leaders for all cliques: {end - start} ms")

    start = current_time_in_millis()
    graph = consolidate_edges(target_graph, clique_graph, leader_annotation)
    end = current_time_in_millis()
    log.info(f"Total time taken to consolidate edges in target graph: {end - start} ms")
    return graph, clique_graph


def build_cliques(target_graph: BaseGraph) -> nx.MultiDiGraph:
    """
    Builds a clique graph from ``same_as`` edges in ``target_graph``.

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        An instance of BaseGraph that contains nodes and edges

    Returns
    -------
    networkx.MultiDiGraph
        The clique graph with only ``same_as`` edges

    """
    clique_graph = nx.MultiDiGraph()
    for n, data in target_graph.nodes(data=True):
        if "same_as" in data:
            new_data = copy.deepcopy(data)
            del new_data["same_as"]
            clique_graph.add_node(n, **new_data)
            for s in data["same_as"]:
                edge_data1 = {"subject": n, "predicate": SAME_AS, "object": s}
                if "provided_by" in data:
                    edge_data1["provided_by"] = data["provided_by"]
                clique_graph.add_edge(n, s, **edge_data1)
                edge_data2 = {"subject": s, "predicate": SAME_AS, "object": n}
                if "provided_by" in data:
                    edge_data2["provided_by"] = data["provided_by"]
                clique_graph.add_edge(s, n, **edge_data2)
    for u, v, data in target_graph.edges(data=True):
        if "predicate" in data and data["predicate"] == SAME_AS:
            # load all biolink:same_as edges to clique_graph
            clique_graph.add_node(u, **target_graph.nodes()[u])
            clique_graph.add_node(v, **target_graph.nodes()[v])
            clique_graph.add_edge(u, v, **data)
            clique_graph.add_edge(
                v,
                u,
                **{
                    "subject": v,
                    "predicate": data["predicate"],
                    "object": v,
                    "relation": data["relation"],
                },
            )
    return clique_graph


def elect_leader(
    target_graph: BaseGraph,
    clique_graph: nx.MultiDiGraph,
    leader_annotation: str,
    prefix_prioritization_map: Optional[Dict[str, List[str]]],
    category_mapping: Optional[Dict[str, str]],
    strict: bool = True,
) -> BaseGraph:
    """
    Elect leader for each clique in a graph.

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        The original graph
    clique_graph: networkx.Graph
        The clique graph
    leader_annotation: str
        The field on a node that signifies that the node is the leader of a clique
    prefix_prioritization_map: Optional[Dict[str, List[str]]]
        A map that gives a prefix priority for one or more categories
    category_mapping: Optional[Dict[str, str]]
        Mapping for non-Biolink Model categories to Biolink Model categories
    strict: bool
        Whether or not to merge nodes in a clique that have conflicting node categories

    Returns
    -------
    kgx.graph.base_graph.BaseGraph
        The updated target graph

    """
    cliques = list(nx.strongly_connected_components(clique_graph))
    log.info(f"Total cliques in clique graph: {len(cliques)}")
    count = 0
    update_dict = {}
    for clique in cliques:
        log.info(
            f"Processing clique: {clique} with {[clique_graph.nodes()[x]['category'] if 'category' in clique_graph.nodes()[x] else None for x in clique]}"
        )
        update_node_categories(
            target_graph, clique_graph, clique, category_mapping, strict
        )
        clique_category, clique_category_ancestors = get_clique_category(
            clique_graph, clique
        )
        log.debug(f"Clique category: {clique_category}")
        invalid_nodes = set()
        for n in clique:
            data = clique_graph.nodes()[n]
            if "_excluded_from_clique" in data and data["_excluded_from_clique"]:
                log.info(
                    f"Removing invalid node {n} from clique graph; node marked to be excluded"
                )
                clique_graph.remove_node(n)
                invalid_nodes.add(n)
            if data["category"][0] not in clique_category_ancestors:
                log.info(
                    f"Removing invalid node {n} from the clique graph; node category {data['category'][0]} not in CCA: {clique_category_ancestors}"
                )
                clique_graph.remove_node(n)
                invalid_nodes.add(n)

        filtered_clique = [x for x in clique if x not in invalid_nodes]
        if filtered_clique:
            if clique_category:
                # First check for LEADER_ANNOTATION property
                leader, election_strategy = get_leader_by_annotation(
                    target_graph, clique_graph, filtered_clique, leader_annotation
                )
                if not leader:
                    # Leader is None; use prefix prioritization strategy
                    log.debug(
                        "Could not elect clique leader by looking for LEADER_ANNOTATION property; "
                        "Using prefix prioritization instead"
                    )
                    if (
                        prefix_prioritization_map
                        and clique_category in prefix_prioritization_map.keys()
                    ):
                        leader, election_strategy = get_leader_by_prefix_priority(
                            target_graph,
                            clique_graph,
                            filtered_clique,
                            prefix_prioritization_map[clique_category],
                        )
                    else:
                        log.debug(
                            f"No prefix order found for category '{clique_category}' in PREFIX_PRIORITIZATION_MAP"
                        )

                if not leader:
                    # Leader is None; fall back to alphabetical sort on prefixes
                    log.debug(
                        "Could not elect clique leader by PREFIX_PRIORITIZATION; Using alphabetical sort on prefixes"
                    )
                    leader, election_strategy = get_leader_by_sort(
                        target_graph, clique_graph, filtered_clique
                    )

                log.debug(
                    f"Elected {leader} as leader via {election_strategy} for clique {filtered_clique}"
                )
                update_dict[leader] = {
                    LEADER_ANNOTATION: True,
                    "election_strategy": election_strategy,
                }
                count += 1

    nx.set_node_attributes(clique_graph, update_dict)
    target_graph.set_node_attributes(target_graph, update_dict)
    log.info(f"Total merged cliques: {count}")
    return target_graph


def consolidate_edges(
    target_graph: BaseGraph, clique_graph: nx.MultiDiGraph, leader_annotation: str
) -> BaseGraph:
    """
    Move all edges from nodes in a clique to the clique leader.

    Original subject and object of a node are preserved via ``ORIGINAL_SUBJECT_PROPERTY`` and ``ORIGINAL_OBJECT_PROPERTY``

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        The original graph
    clique_graph: networkx.MultiDiGraph
        The clique graph
    leader_annotation: str
        The field on a node that signifies that the node is the leader of a clique

    Returns
    -------
    kgx.graph.base_graph.BaseGraph
        The target graph where all edges from nodes in a clique are moved to clique leader

    """
    cliques = list(nx.strongly_connected_components(clique_graph))
    log.info(f"Consolidating edges in {len(cliques)} cliques")
    for clique in cliques:
        log.debug(f"Processing clique: {clique}")
        leaders: List = [
            x
            for x in clique
            if leader_annotation in clique_graph.nodes()[x]
            and clique_graph.nodes()[x][leader_annotation]
        ]
        if len(leaders) == 0:
            log.debug("No leader elected for clique {}; skipping".format(clique))
            continue
        leader: str = leaders[0]
        # update nodes in target graph
        target_graph.set_node_attributes(
            target_graph,
            {
                leader: {
                    leader_annotation: clique_graph.nodes()[leader].get(
                        leader_annotation
                    ),
                    "election_strategy": clique_graph.nodes()[leader].get(
                        "election_strategy"
                    ),
                }
            },
        )
        leader_equivalent_identifiers = set([x for x in clique_graph.neighbors(leader)])
        for node in clique:
            if node == leader:
                continue
            log.debug(f"Looking for in_edges for {node}")
            in_edges = target_graph.in_edges(node, keys=False, data=True)
            filtered_in_edges = [x for x in in_edges if x[2]["predicate"] != SAME_AS]
            equiv_in_edges = [x for x in in_edges if x[2]["predicate"] == SAME_AS]
            log.debug(f"Moving {len(in_edges)} in-edges from {node} to {leader}")
            for u, v, edge_data in filtered_in_edges:
                key = generate_edge_key(u, edge_data["predicate"], v)
                target_graph.remove_edge(u, v, edge_key=key)
                edge_data[ORIGINAL_SUBJECT_PROPERTY] = edge_data["subject"]
                edge_data[ORIGINAL_OBJECT_PROPERTY] = edge_data["object"]
                edge_data["object"] = leader
                key = generate_edge_key(u, edge_data["predicate"], leader)
                if (
                    edge_data["subject"] == edge_data["object"]
                    and edge_data["predicate"] == SUBCLASS_OF
                ):
                    continue
                target_graph.add_edge(
                    edge_data["subject"], edge_data["object"], key, **edge_data
                )

            log.debug(f"Looking for out_edges for {node}")
            out_edges = target_graph.out_edges(node, keys=False, data=True)
            filtered_out_edges = [x for x in out_edges if x[2]["predicate"] != SAME_AS]
            equiv_out_edges = [x for x in out_edges if x[2]["predicate"] == SAME_AS]
            log.debug(f"Moving {len(out_edges)} out-edges from {node} to {leader}")
            for u, v, edge_data in filtered_out_edges:
                key = generate_edge_key(u, edge_data["predicate"], v)
                target_graph.remove_edge(u, v, edge_key=key)
                edge_data[ORIGINAL_SUBJECT_PROPERTY] = edge_data["subject"]
                edge_data[ORIGINAL_OBJECT_PROPERTY] = edge_data["object"]
                edge_data["subject"] = leader
                key = generate_edge_key(leader, edge_data["predicate"], v)
                if (
                    edge_data["subject"] == edge_data["object"]
                    and edge_data["predicate"] == SUBCLASS_OF
                ):
                    continue
                target_graph.add_edge(
                    edge_data["subject"], edge_data["object"], key, **edge_data
                )

            log.debug(f"equiv out edges: {equiv_out_edges}")
            equivalent_identifiers = set()
            for u, v, edge_data in equiv_in_edges:
                if u != leader:
                    equivalent_identifiers.add(u)
                if v != leader:
                    equivalent_identifiers.add(v)
                target_graph.remove_edge(
                    u, v, edge_key=generate_edge_key(u, SAME_AS, v)
                )

            log.debug(f"equiv out edges: {equiv_out_edges}")
            for u, v, edge_data in equiv_out_edges:
                if u != leader:
                    log.debug(f"{u} is an equivalent identifier of leader {leader}")
                    equivalent_identifiers.add(u)
                if v != leader:
                    log.debug(f"{v} is an equivalent identifier of leader {leader}")
                    equivalent_identifiers.add(v)
                target_graph.remove_edge(
                    u, v, edge_key=generate_edge_key(u, SAME_AS, v)
                )

            leader_equivalent_identifiers.update(equivalent_identifiers)

        log.debug(
            f"setting same_as property to leader node with {leader_equivalent_identifiers}"
        )
        target_graph.set_node_attributes(
            target_graph, {leader: {"same_as": list(leader_equivalent_identifiers)}}
        )
        log.debug(
            f"removing equivalent nodes of leader: {leader_equivalent_identifiers}"
        )
        for n in leader_equivalent_identifiers:
            target_graph.remove_node(n)
    return target_graph


def update_node_categories(
    target_graph: BaseGraph,
    clique_graph: nx.MultiDiGraph,
    clique: List,
    category_mapping: Optional[Dict[str, str]],
    strict: bool = True,
) -> List:
    """
    For a given clique, get category for each node in clique and validate against Biolink Model,
    mapping to Biolink Model category where needed.

    For example, If a node has ``biolink:Gene`` as its category, then this method adds all of its ancestors.

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        The original graph
    clique_graph: networkx.Graph
        The clique graph
    clique: List
        A list of nodes from a clique
    category_mapping: Optional[Dict[str, str]]
        Mapping for non-Biolink Model categories to Biolink Model categories
    strict: bool
        Whether or not to merge nodes in a clique that have conflicting node categories

    Returns
    -------
    List
        The clique

    """
    updated_clique_graph_properties = {}
    updated_target_graph_properties = {}
    for node in clique:
        # For each node in a clique, get its category property
        data = clique_graph.nodes()[node]
        if "category" in data:
            categories = data["category"]
        else:
            categories = get_category_from_equivalence(
                target_graph, clique_graph, node, data
            )

        # differentiate between valid and invalid categories
        (
            valid_biolink_categories,
            invalid_biolink_categories,
            invalid_categories,
        ) = check_all_categories(categories)
        log.debug(
            f"valid biolink categories: {valid_biolink_categories} invalid biolink categories: {invalid_biolink_categories} invalid_categories: {invalid_categories}"
        )
        # extend categories to have the longest list of ancestors
        extended_categories: List = []
        for x in valid_biolink_categories:
            ancestors = get_biolink_ancestors(x)
            if len(ancestors) > len(extended_categories):
                extended_categories.extend(ancestors)
        log.debug(f"Extended categories: {extended_categories}")
        clique_graph_update_dict: Dict = {"category": list(extended_categories)}
        target_graph_update_dict: Dict = {}

        if invalid_biolink_categories:
            if strict:
                clique_graph_update_dict["_excluded_from_clique"] = True
                target_graph_update_dict["_excluded_from_clique"] = True
            clique_graph_update_dict[
                "invalid_biolink_category"
            ] = invalid_biolink_categories
            target_graph_update_dict[
                "invalid_biolink_category"
            ] = invalid_biolink_categories

        if invalid_categories:
            clique_graph_update_dict["_invalid_category"] = invalid_categories
            target_graph_update_dict["_invalid_category"] = invalid_categories

        updated_clique_graph_properties[node] = clique_graph_update_dict
        updated_target_graph_properties[node] = target_graph_update_dict

    nx.set_node_attributes(clique_graph, updated_clique_graph_properties)
    target_graph.set_node_attributes(target_graph, updated_target_graph_properties)
    return clique


def get_clique_category(
    clique_graph: nx.MultiDiGraph, clique: List
) -> Tuple[str, List]:
    """
    Given a clique, identify the category of the clique.

    Parameters
    ----------
    clique_graph: nx.MultiDiGraph
        Clique graph
    clique: List
        A list of nodes in clique

    Returns
    -------
    Tuple[str, list]
        A tuple of clique category and its ancestors

    """
    l = [clique_graph.nodes()[x]["category"] for x in clique]
    u = OrderedSet.union(*l)
    uo = sort_categories(u)
    log.debug(f"outcome of union (sorted): {uo}")
    clique_category = uo[0]
    clique_category_ancestors = get_biolink_ancestors(uo[0])
    return clique_category, clique_category_ancestors


def check_categories(
    categories: List, closure: List, category_mapping: Optional[Dict[str, str]] = None
) -> Tuple[List, List, List]:
    """
    Check categories to ensure whether values in ``categories`` are valid biolink categories.
    Valid biolink categories are classes that descend from 'NamedThing'.
    Mixins, while valid ancestors, are not valid categories.

    Parameters
    ----------
    categories: List
        A list of categories to check
    closure: List
        A list of nodes in a clique
    category_mapping: Optional[Dict[str, str]]
        A map that provides mapping from a non-biolink category to a biolink category

    Returns
    -------
    Tuple[List, List, List]
        A tuple consisting of valid biolink categories, invalid biolink categories, and invalid categories

    """
    valid_biolink_categories = []
    invalid_biolink_categories = []
    invalid_categories = []
    tk = get_toolkit()
    for x in categories:
        # use the toolkit to check if the declared category is actually a mixin.
        if tk.is_mixin(x):
            invalid_categories.append(x)
            continue
        # get biolink element corresponding to category
        element = get_biolink_element(x)
        if element:
            mapped_category = format_biolink_category(element["name"])
            if mapped_category in closure:
                valid_biolink_categories.append(x)
            else:
                log.warning(f"category '{mapped_category}' not in closure: {closure}")
                if category_mapping:
                    mapped = category_mapping[x] if x in category_mapping.keys() else x
                    if mapped not in closure:
                        log.warning(
                            f"category '{mapped_category}' is not in category_mapping."
                        )
                        invalid_biolink_categories.append(x)
                else:
                    invalid_biolink_categories.append(x)
        else:
            log.warning(f"category '{x}' is not in Biolink Model")
            invalid_categories.append(x)
            continue
    return valid_biolink_categories, invalid_biolink_categories, invalid_categories


def check_all_categories(categories) -> Tuple[List, List, List]:
    """
    Check all categories in ``categories``.

    Parameters
    ----------
    categories: List
        A list of categories

    Returns
    -------
    Tuple[List, List, List]
        A tuple consisting of valid biolink categories, invalid biolink categories, and invalid categories

    Note: the sort_categories method will re-arrange the passed in category list according to the distance
    of each list member from the top of their hierarchy.  Each category's hierarchy is made up of its
    'is_a' and mixin ancestors.

    """
    previous: List = []
    valid_biolink_categories: List = []
    invalid_biolink_categories: List = []
    invalid_categories: List = []
    sc: List = sort_categories(categories)
    for c in sc:
        if previous:
            vbc, ibc, ic = check_categories(
                [c], get_biolink_ancestors(previous[0]), None
            )
        else:
            vbc, ibc, ic = check_categories([c], get_biolink_ancestors(c), None)
        if vbc:
            valid_biolink_categories.extend(vbc)
        if ic:
            invalid_categories.extend(ic)
        if ibc:
            invalid_biolink_categories.extend(ibc)
        else:
            previous = vbc
    return valid_biolink_categories, invalid_biolink_categories, invalid_categories


def sort_categories(categories: Union[List, Set, OrderedSet]) -> List:
    """
    Sort a list of categories from most specific to the most generic.

    Parameters
    ----------
    categories: Union[List, Set, OrderedSet]
        A list of categories

    Returns
    -------
    List
        A sorted list of categories where sorted means that the first element in the list returned
        has the most number of parents in the class hierarchy.

    """
    weighted_categories = []
    for c in categories:
        weighted_categories.append((len(get_biolink_ancestors(c)), c))
    sorted_categories = sorted(weighted_categories, key=lambda x: x[0], reverse=True)
    return [x[1] for x in sorted_categories]


def get_category_from_equivalence(
    target_graph: BaseGraph, clique_graph: nx.MultiDiGraph, node: str, attributes: Dict
) -> List:
    """
    Get category for a node based on its equivalent nodes in a graph.

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        The original graph
    clique_graph: networkx.MultiDiGraph
        The clique graph
    node: str
        Node identifier
    attributes: Dict
        Node's attributes

    Returns
    -------
    List
        Category for the node

    """
    category: List = []
    for u, v, data in clique_graph.edges(node, data=True):
        if data["predicate"] == SAME_AS:
            if u == node:
                if "category" in clique_graph.nodes()[v]:
                    category = clique_graph.nodes()[v]["category"]
                    break
            elif v == node:
                if "category" in clique_graph.nodes()[u]:
                    category = clique_graph.nodes()[u]["category"]
                    break
            update = {node: {"category": category}}
            nx.set_node_attributes(clique_graph, update)
    return category


def get_leader_by_annotation(
    target_graph: BaseGraph,
    clique_graph: nx.MultiDiGraph,
    clique: List,
    leader_annotation: str,
) -> Tuple[Optional[str], Optional[str]]:
    """
    Get leader by searching for leader annotation property in any of the nodes in a given clique.

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        The original graph
    clique_graph: networkx.MultiDiGraph
        The clique graph
    clique: List
        A list of nodes from a clique
    leader_annotation: str
        The field on a node that signifies that the node is the leader of a clique

    Returns
    -------
    Tuple[Optional[str], Optional[str]]
        A tuple containing the node that has been elected as the leader and the election strategy

    """
    leader = None
    election_strategy = None
    for node in clique:
        attributes = clique_graph.nodes()[node]
        if leader_annotation in attributes:
            if isinstance(attributes[leader_annotation], str):
                v = attributes[leader_annotation]
                if v == "true" or v == "True":
                    leader = node
            elif isinstance(attributes[leader_annotation], list):
                v = attributes[leader_annotation][0]
                if isinstance(v, str):
                    if v == "true" or v == "True":
                        leader = node
                elif isinstance(v, bool):
                    if eval(str(v)):
                        leader = node
            elif isinstance(attributes[leader_annotation], bool):
                v = attributes[leader_annotation]
                if eval(str(v)):
                    leader = node
    if leader:
        election_strategy = "LEADER_ANNOTATION"
        log.debug(f"Elected leader '{leader}' via LEADER_ANNOTATION")

    return leader, election_strategy


def get_leader_by_prefix_priority(
    target_graph: BaseGraph,
    clique_graph: nx.MultiDiGraph,
    clique: List,
    prefix_priority_list: List,
) -> Tuple[Optional[str], Optional[str]]:
    """
    Get leader from clique based on a given prefix priority.

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        The original graph
    clique_graph: networkx.MultiDiGraph
        The clique graph
    clique: List
        A list of nodes that correspond to a clique
    prefix_priority_list: List
        A list of prefixes in descending priority

    Returns
    -------
    Tuple[Optional[str], Optional[str]]
        A tuple containing the node that has been elected as the leader and the election strategy

    """
    leader = None
    election_strategy = None
    for prefix in prefix_priority_list:
        log.debug(f"Checking for prefix {prefix} in {clique}")
        leader = next((x for x in clique if prefix in x), None)
        if leader:
            election_strategy = "PREFIX_PRIORITIZATION"
            log.debug(f"Elected leader '{leader}' via {election_strategy}")
            break
    return leader, election_strategy


def get_leader_by_sort(
    target_graph: BaseGraph, clique_graph: nx.MultiDiGraph, clique: List
) -> Tuple[Optional[str], Optional[str]]:
    """
    Get leader from clique based on the first selection from an alphabetical sort of the node id prefixes.

    Parameters
    ----------
    target_graph: kgx.graph.base_graph.BaseGraph
        The original graph
    clique_graph: networkx.MultiDiGraph
        The clique graph
    clique: List
        A list of nodes that correspond to a clique

    Returns
    -------
    Tuple[Optional[str], Optional[str]]
        A tuple containing the node that has been elected as the leader and the election strategy

    """
    election_strategy = "ALPHABETICAL_SORT"
    prefixes = [x.split(":", 1)[0] for x in clique]
    prefixes.sort()
    leader_prefix = prefixes[0]
    log.debug(f"clique: {clique} leader_prefix: {leader_prefix}")
    leader = [x for x in clique if leader_prefix in x]
    if leader:
        log.debug(f"Elected leader '{leader}' via {election_strategy}")
    return leader[0], election_strategy



from typing import Set, Optional, Generator, Any, Dict

import rdflib
from rdflib import Namespace, URIRef, OWL, RDFS, RDF

from kgx.config import get_logger
from kgx.source import RdfSource
from kgx.utils.kgx_utils import (
    current_time_in_millis,
    generate_uuid,
    generate_edge_identifiers,
    validate_node,
    sanitize_import,
    validate_edge,
)

log = get_logger()


class OwlSource(RdfSource):
    """
    OwlSource is responsible for parsing an OWL ontology.

    ..note::
        This is a simple parser that loads direct class-class relationships.
        For more formal OWL parsing, refer to Robot: http://robot.obolibrary.org/

    """

    def __init__(self):
        self.imported: Set = set()
        super().__init__()
        self.OWLSTAR = Namespace("http://w3id.org/owlstar/")
        self.excluded_predicates = {
            URIRef("https://raw.githubusercontent.com/geneontology/go-ontology/master/contrib/oboInOwl#id")
        }

    def parse(
        self,
        filename: str,
        format: str = "owl",
        compression: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from an OWL and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``owl``)
        compression: Optional[str]
            The compression type (``gz``)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records read from the file

        """
        rdfgraph = rdflib.Graph()
        if compression:
            log.warning(f"compression mode '{compression}' not supported by OwlSource")
        if format is None:
            format = rdflib.util.guess_format(filename)

        if format == "owl":
            format = "xml"

        log.info("Parsing {} with '{}' format".format(filename, format))
        rdfgraph.parse(filename, format=format)
        log.info("{} parsed with {} triples".format(filename, len(rdfgraph)))

        self.set_provenance_map(kwargs)

        self.start = current_time_in_millis()
        log.info(f"Done parsing {filename}")

        triples = rdfgraph.triples((None, OWL.imports, None))
        for s, p, o in triples:
            # Load all imports first
            if p == OWL.imports:
                if o not in self.imported:
                    input_format = rdflib.util.guess_format(o)
                    imported_rdfgraph = rdflib.Graph()
                    log.info(f"Parsing OWL import: {o}")
                    self.imported.add(o)
                    imported_rdfgraph.parse(o, format=input_format)
                    self.load_graph(imported_rdfgraph)
                else:
                    log.warning(f"Trying to import {o} but its already done")
        yield from self.load_graph(rdfgraph)

    def load_graph(self, rdfgraph: rdflib.Graph, **kwargs: Any) -> None:
        """
        Walk through the rdflib.Graph and load all triples into kgx.graph.base_graph.BaseGraph

        Parameters
        ----------
        rdfgraph: rdflib.Graph
            Graph containing nodes and edges
        kwargs: Any
            Any additional arguments

        """
        seen = set()
        seen.add(RDFS.subClassOf)
        for s, p, o in rdfgraph.triples((None, RDFS.subClassOf, None)):
            # ignoring blank nodes
            if isinstance(s, rdflib.term.BNode):
                continue
            pred = None
            parent = None
            os_interpretation = None
            if isinstance(o, rdflib.term.BNode):
                # C SubClassOf R some D
                for x in rdfgraph.objects(o, OWL.onProperty):
                    pred = x
                # owl:someValuesFrom
                for x in rdfgraph.objects(o, OWL.someValuesFrom):
                    os_interpretation = self.OWLSTAR.term("AllSomeInterpretation")
                    parent = x
                # owl:allValuesFrom
                for x in rdfgraph.objects(o, OWL.allValuesFrom):
                    os_interpretation = self.OWLSTAR.term("AllOnlyInterpretation")
                    parent = x
                if pred is None or parent is None:
                    log.warning(
                        f"{s} {p} {o} has OWL.onProperty {pred} and OWL.someValuesFrom {parent}"
                    )
                    log.warning("Do not know how to handle BNode: {}".format(o))
                    continue
            else:
                # C rdfs:subClassOf D (where C and D are named classes)
                pred = p
                parent = o
            if os_interpretation:
                # reify edges that have logical interpretation
                eid = generate_uuid()
                self.reified_nodes.add(eid)
                yield from self.triple(
                    URIRef(eid), self.BIOLINK.term("category"), self.BIOLINK.Association
                )
                yield from self.triple(URIRef(eid), self.BIOLINK.term("subject"), s)
                yield from self.triple(
                    URIRef(eid), self.BIOLINK.term("predicate"), pred
                )
                yield from self.triple(URIRef(eid), self.BIOLINK.term("object"), parent)
                yield from self.triple(
                    URIRef(eid),
                    self.BIOLINK.term("logical_interpretation"),
                    os_interpretation,
                )
            else:
                yield from self.triple(s, pred, parent)

        seen.add(OWL.equivalentClass)
        for s, p, o in rdfgraph.triples((None, OWL.equivalentClass, None)):
            # A owl:equivalentClass B (where A and B are named classes)
            if not isinstance(o, rdflib.term.BNode):
                yield from self.triple(s, p, o)

        for relation in rdfgraph.subjects(RDF.type, OWL.ObjectProperty):
            seen.add(relation)
            for s, p, o in rdfgraph.triples((relation, None, None)):
                if not isinstance(o, rdflib.term.BNode):
                    if p not in self.excluded_predicates:
                        yield from self.triple(s, p, o)

        for s, p, o in rdfgraph.triples((None, None, None)):
            if isinstance(s, rdflib.term.BNode) or isinstance(o, rdflib.term.BNode):
                continue
            if p in seen:
                continue
            if p in self.excluded_predicates:
                continue
            yield from self.triple(s, p, o)

        for n in self.reified_nodes:
            data = self.node_cache.pop(n)
            self.dereify(n, data)

        for k, data in self.node_cache.items():
            node_data = validate_node(data)
            node_data = sanitize_import(node_data)
            self.set_node_provenance(node_data)
            if self.check_node_filter(node_data):
                self.node_properties.update(node_data.keys())
                yield k, node_data
        self.node_cache.clear()

        for k, data in self.edge_cache.items():
            edge_data = validate_edge(data)
            edge_data = sanitize_import(edge_data)
            self.set_edge_provenance(edge_data)
            if self.check_edge_filter(edge_data):
                self.edge_properties.update(edge_data.keys())
                yield k[0], k[1], k[2], edge_data
        self.edge_cache.clear()



import kgx
import click
from typing import List, Tuple, Optional, Set

from kgx.config import get_logger, get_config
from kgx.cli.cli_utils import (
    get_input_file_types,
    parse_source,
    apply_operations,
    graph_summary,
    validate,
    neo4j_download,
    neo4j_upload,
    transform,
    merge,
    summary_report_types,
    get_report_format_types,
)

log = get_logger()
config = get_config()


def error(msg):
    log.error(msg)
    quit()


@click.group()
@click.version_option(version=kgx.__version__, prog_name=kgx.__name__)
def cli():
    """
    Knowledge Graph Exchange CLI entrypoint.
    \f

    """
    pass


@cli.command(name="graph-summary")
@click.argument("inputs", required=True, type=click.Path(exists=True), nargs=-1)
@click.option(
    "--input-format",
    "-i",
    required=True,
    help=f"The input format. Can be one of {get_input_file_types()}",
)
@click.option(
    "--input-compression", "-c", required=False, help="The input compression type"
)
@click.option("--output", "-o", required=True, type=click.Path(exists=False))
@click.option(
    "--report-type",
    "-r",
    required=False,
    type=str,
    help=f"The summary report type. Must be one of {tuple(summary_report_types.keys())}",
    default="kgx-map",
)
@click.option(
    "--report-format",
    "-f",
    help=f"The input format. Can be one of {get_report_format_types()}",
)
@click.option("--stream", "-s", is_flag=True, help="Parse input as a stream")
@click.option(
    "--graph-name",
    "-n",
    required=False,
    help="User specified name of graph being summarized (default: 'Graph')",
)
@click.option(
    "--node-facet-properties",
    required=False,
    multiple=True,
    help="A list of node properties from which to generate counts per value for those properties",
)
@click.option(
    "--edge-facet-properties",
    required=False,
    multiple=True,
    help="A list of edge properties from which to generate counts per value for those properties",
)
@click.option(
    "--error-log",
    "-l",
    required=False,
    type=click.Path(exists=False),
    help='File within which to report graph data parsing errors (default: "stderr")',
)
def graph_summary_wrapper(
    inputs: List[str],
    input_format: str,
    input_compression: Optional[str],
    output: Optional[str],
    report_type: str,
    report_format: str,
    stream: bool,
    graph_name: str,
    node_facet_properties: Optional[List],
    edge_facet_properties: Optional[List],
    error_log: str = ''
):
    """
    Loads and summarizes a knowledge graph from a set of input files.
    \f

    Parameters
    ----------
    inputs: List[str]
        Input file
    input_format: str
        Input file format
    input_compression: Optional[str]
        The input compression type
    output: Optional[str]
        Where to write the output (stdout, by default)
    report_type: str
        The summary report type
    report_format: Optional[str]
        The summary report format file types: 'yaml' or 'json'  (default is report_type specific)
    stream: bool
        Whether to parse input as a stream
    graph_name: str
        User specified name of graph being summarize
    node_facet_properties: Optional[List]
        A list of node properties from which to generate counts per value for those properties.
        For example, ``['provided_by']``
    edge_facet_properties: Optional[List]
        A list of edge properties from which to generate counts per value for those properties.
        For example, ``['original_knowledge_source', 'aggregator_knowledge_source']``
    error_log: str
        Where to write any graph processing error message (stderr, by default, for empty argument)
    """
    graph_summary(
        inputs,
        input_format,
        input_compression,
        output,
        report_type,
        report_format,
        stream,
        graph_name,
        node_facet_properties=list(node_facet_properties),
        edge_facet_properties=list(edge_facet_properties),
        error_log=error_log,
    )


@cli.command(name="validate")
@click.argument("inputs", required=True, type=click.Path(exists=True), nargs=-1)
@click.option(
    "--input-format",
    "-i",
    required=True,
    help=f"The input format. Can be one of {get_input_file_types()}",
)
@click.option(
    "--input-compression", "-c", required=False, help="The input compression type"
)
@click.option(
    "--output",
    "-o",
    required=False,
    type=click.Path(exists=False),
    help="File to write validation reports to",
)
@click.option("--stream", "-s", is_flag=True, help="Parse input as a stream")
@click.option(
    "--biolink-release",
    "-b",
    required=False,
    help="Biolink Model Release (SemVer) used for validation (default: latest Biolink Model Toolkit version)",
)
def validate_wrapper(
    inputs: List[str],
    input_format: str,
    input_compression: str,
    output: str,
    stream: bool,
    biolink_release: str = None,
):
    """
    Run KGX validator on an input file to check for Biolink Model compliance.
    \f

    Parameters
    ----------
    inputs: List[str]
        Input files
    input_format: str
        The input format
    input_compression: str
        The input compression type
    output: str
        Path to output file
    stream: bool
        Whether to parse input as a stream
    biolink_release: Optional[str]
        SemVer version of Biolink Model Release used for validation (default: latest Biolink Model Toolkit version)
    """
    validate(inputs, input_format, input_compression, output, stream, biolink_release)


@cli.command(name="neo4j-download")
@click.option(
    "--uri",
    "-l",
    required=True,
    type=str,
    help="Neo4j URI to download from. For example, https://localhost:7474",
)
@click.option("--username", "-u", required=True, type=str, help="Neo4j username")
@click.option("--password", "-p", required=True, type=str, help="Neo4j password")
@click.option(
    "--output", "-o", required=True, type=click.Path(exists=False), help="Output"
)
@click.option(
    "--output-format",
    "-f",
    required=True,
    help=f"The output format. Can be one of {get_input_file_types()}",
)
@click.option(
    "--output-compression", "-d", required=False, help="The output compression type"
)
@click.option("--stream", "-s", is_flag=True, help="Parse input as a stream")
@click.option(
    "--node-filters",
    "-n",
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f"Filters for filtering nodes from the input graph",
)
@click.option(
    "--edge-filters",
    "-e",
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f"Filters for filtering edges from the input graph",
)
def neo4j_download_wrapper(
    uri: str,
    username: str,
    password: str,
    output: str,
    output_format: str,
    output_compression: str,
    stream: bool,
    node_filters: Tuple,
    edge_filters: Tuple,
):
    """
    Download nodes and edges from Neo4j database.
    \f

    Parameters
    ----------
    uri: str
        Neo4j URI. For example, https://localhost:7474
    username: str
        Username for authentication
    password: str
        Password for authentication
    output: str
        Where to write the output (stdout, by default)
    output_format: str
        The output type (``tsv``, by default)
    output_compression: str
        The output compression type
    stream: bool
        Whether to parse input as a stream
    node_filters: Tuple[str, str]
        Node filters
    edge_filters: Tuple[str, str]
        Edge filters

    """
    neo4j_download(
        uri,
        username,
        password,
        output,
        output_format,
        output_compression,
        stream,
        node_filters,
        edge_filters,
    )


@cli.command(name="neo4j-upload")
@click.argument("inputs", required=True, type=click.Path(exists=True), nargs=-1)
@click.option(
    "--input-format",
    "-i",
    required=True,
    help=f"The input format. Can be one of {get_input_file_types()}",
)
@click.option(
    "--input-compression", "-c", required=False, help="The input compression type"
)
@click.option(
    "--uri",
    "-l",
    required=True,
    type=str,
    help="Neo4j URI to upload to. For example, https://localhost:7474",
)
@click.option("--username", "-u", required=True, type=str, help="Neo4j username")
@click.option("--password", "-p", required=True, type=str, help="Neo4j password")
@click.option("--stream", "-s", is_flag=True, help="Parse input as a stream")
@click.option(
    "--node-filters",
    "-n",
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f"Filters for filtering nodes from the input graph",
)
@click.option(
    "--edge-filters",
    "-e",
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f"Filters for filtering edges from the input graph",
)
def neo4j_upload_wrapper(
    inputs: List[str],
    input_format: str,
    input_compression: str,
    uri: str,
    username: str,
    password: str,
    stream: bool,
    node_filters: Tuple[str, str],
    edge_filters: Tuple[str, str],
):
    """
    Upload a set of nodes/edges to a Neo4j database.
    \f

    Parameters
    ----------
    inputs: List[str]
        A list of files that contains nodes/edges
    input_format: str
        The input format
    input_compression: str
        The input compression type
    uri: str
        The full HTTP address for Neo4j database
    username: str
        Username for authentication
    password: str
        Password for authentication
    stream: bool
        Whether to parse input as a stream
    node_filters: Tuple[str, str]
        Node filters
    edge_filters: Tuple[str, str]
        Edge filters

    """
    neo4j_upload(
        inputs,
        input_format,
        input_compression,
        uri,
        username,
        password,
        stream,
        node_filters,
        edge_filters,
    )


@cli.command(name="transform")
@click.argument("inputs", required=False, type=click.Path(exists=True), nargs=-1)
@click.option(
    "--input-format",
    "-i",
    required=False,
    help=f"The input format. Can be one of {get_input_file_types()}",
)
@click.option(
    "--input-compression", "-c", required=False, help="The input compression type"
)
@click.option(
    "--output", "-o", required=False, type=click.Path(exists=False), help="Output"
)
@click.option(
    "--output-format",
    "-f",
    required=False,
    help=f"The output format. Can be one of {get_input_file_types()}",
)
@click.option(
    "--output-compression", "-d", required=False, help="The output compression type"
)
@click.option("--stream", is_flag=True, help="Parse input as a stream")
@click.option(
    "--node-filters",
    "-n",
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f"Filters for filtering nodes from the input graph",
)
@click.option(
    "--edge-filters",
    "-e",
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help=f"Filters for filtering edges from the input graph",
)
@click.option(
    "--transform-config", required=False, type=str, help=f"Transform config YAML"
)
@click.option(
    "--source",
    required=False,
    type=str,
    multiple=True,
    help="Source(s) from the YAML to process",
)
@click.option(
    "--knowledge-sources",
    "-k",
    required=False,
    type=click.Tuple([str, str]),
    multiple=True,
    help="A named knowledge source with (string, boolean or tuple rewrite) specification",
)
@click.option(
    "--infores-catalog",
    required=False,
    type=click.Path(exists=False),
    help="Optional dump of a CSV file of InfoRes CURIE to Knowledge Source mappings",
)
@click.option(
    "--processes",
    "-p",
    required=False,
    type=int,
    default=1,
    help="Number of processes to use",
)
def transform_wrapper(
    inputs: List[str],
    input_format: str,
    input_compression: str,
    output: str,
    output_format: str,
    output_compression: str,
    stream: bool,
    node_filters: Optional[List[Tuple[str, str]]],
    edge_filters: Optional[List[Tuple[str, str]]],
    transform_config: str,
    source: Optional[List],
    knowledge_sources: Optional[List[Tuple[str, str]]],
    processes: int,
    infores_catalog: Optional[str] = None,
):
    """
    Transform a Knowledge Graph from one serialization form to another.
    \f

    Parameters
    ----------
    inputs: List[str]
        A list of files that contains nodes/edges
    input_format: str
        The input format
    input_compression: str
        The input compression type
    output: str
        The output file
    output_format: str
        The output format
    output_compression: str
        The output compression typ
    stream: bool
        Whether or not to stream
    node_filters: Optional[List[Tuple[str, str]]]
        Node input filters
    edge_filters: Optional[List[Tuple[str, str]]]
        Edge input filters
    transform_config: str
        Transform config YAML
    source: List
        A list of source(s) to load from the YAML
    knowledge_sources: Optional[List[Tuple[str, str]]]
        A list of named knowledge sources with (string, boolean or tuple rewrite) specification
    infores_catalog: Optional[str]
        Optional dump of a TSV file of InfoRes CURIE to Knowledge Source mappings
    processes: int
        Number of processes to use

    """
    transform(
        inputs,
        input_format=input_format,
        input_compression=input_compression,
        output=output,
        output_format=output_format,
        output_compression=output_compression,
        stream=stream,
        node_filters=node_filters,
        edge_filters=edge_filters,
        transform_config=transform_config,
        source=source,
        knowledge_sources=knowledge_sources,
        processes=processes,
        infores_catalog=infores_catalog,
    )


@cli.command(name="merge")
@click.option("--merge-config", required=True, type=str)
@click.option(
    "--source",
    required=False,
    type=str,
    multiple=True,
    help="Source(s) from the YAML to process",
)
@click.option(
    "--destination",
    required=False,
    type=str,
    multiple=True,
    help="Destination(s) from the YAML to process",
)
@click.option(
    "--processes",
    "-p",
    required=False,
    type=int,
    default=1,
    help="Number of processes to use",
)
def merge_wrapper(merge_config: str, source: List, destination: List, processes: int):
    """
    Load nodes and edges from files and KGs, as defined in a config YAML, and merge them into a single graph.
    The merged graph can then be written to a local/remote Neo4j instance OR be serialized into a file.
    \f

    .. note::
        Everything here is driven by the ``merge-config`` YAML.

    Parameters
    ----------
    merge_config: str
        Merge config YAML
    source: List
        A list of source to load from the YAML
    destination: List
        A list of destination to write to, as defined in the YAML
    processes: int
        Number of processes to use

    """
    merge(merge_config, source, destination, processes)



from typing import Dict

from kgx.graph.base_graph import BaseGraph

from kgx.config import get_graph_store_class
from kgx.sink.sink import Sink
from kgx.utils.kgx_utils import generate_edge_key


class GraphSink(Sink):
    """
    GraphSink is responsible for writing data as records
    to an in memory graph representation.

    The underlying store is determined by the graph store
    class defined in config (``kgx.graph.nx_graph.NxGraph``, by default).

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        An instance of BaseGraph to read from

    """

    def __init__(self, graph: BaseGraph = None):
        super().__init__()
        if graph:
            self.graph = graph
        else:
            self.graph = get_graph_store_class()()

    def write_node(self, record: Dict) -> None:
        """
        Write a node record to graph.

        Parameters
        ----------
        record: Dict
            A node record

        """
        self.graph.add_node(record["id"], **record)

    def write_edge(self, record: Dict) -> None:
        """
        Write an edge record to graph.

        Parameters
        ----------
        record: Dict
            An edge record

        """
        key = (
            record["key"]
            if "key" in record
            else generate_edge_key(
                record["subject"], record["predicate"], record["object"]
            )
        )
        self.graph.add_edge(record["subject"], record["object"], key, **record)

    def finalize(self) -> None:
        """
        Perform any operations after writing nodes and edges to graph.
        """
        pass



from typing import Dict, List, Optional, Any, Callable, Set
from sys import stderr

import re

import yaml
from json import dump
from json.encoder import JSONEncoder

from kgx.utils.kgx_utils import GraphEntityType
from kgx.graph.base_graph import BaseGraph
from kgx.prefix_manager import PrefixManager

TOTAL_NODES = "total_nodes"
NODE_CATEGORIES = "node_categories"

NODE_ID_PREFIXES_BY_CATEGORY = "node_id_prefixes_by_category"
NODE_ID_PREFIXES = "node_id_prefixes"

COUNT_BY_CATEGORY = "count_by_category"

COUNT_BY_ID_PREFIXES_BY_CATEGORY = "count_by_id_prefixes_by_category"
COUNT_BY_ID_PREFIXES = "count_by_id_prefixes"

TOTAL_EDGES = "total_edges"
EDGE_PREDICATES = "predicates"
COUNT_BY_EDGE_PREDICATES = "count_by_predicates"
COUNT_BY_SPO = "count_by_spo"


# Note: the format of the stats generated might change in the future

####################################################################################
# New "Inspector Class" design pattern for KGX stream data processing
####################################################################################


def gs_default(o):
    """
    JSONEncoder 'default' function override to
    properly serialize 'Set' objects (into 'List')
    """
    if isinstance(o, GraphSummary.Category):
        return o.json_object()
    else:
        try:
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)
        # Let the base class default method raise the TypeError
        return JSONEncoder().default(o)


_category_curie_regexp = re.compile("^biolink:[A-Z][a-zA-Z]*$")
_predicate_curie_regexp = re.compile("^biolink:[a-z][a-z_]*$")


# DRY parse warning message
def _parse_warning(prefix: str, item: str, suffix: str = None):
    print(
        prefix + " '" + item + "'" + (" " + suffix if suffix else "") + "? Ignoring...",
        file=GraphSummary.error_log,
    )


class GraphSummary:
    """
    Class for generating a "classical" knowledge graph summary.

    The optional 'progress_monitor' for the validator should be a lightweight Callable
    which is injected into the class 'inspector' Callable, designed to intercepts
    node and edge records streaming through the Validator (inside a Transformer.process() call.
    The first (GraphEntityType) argument of the Callable tags the record as a NODE or an EDGE.
    The second argument given to the Callable is the current record itself.
    This Callable is strictly meant to be procedural and should *not* mutate the record.
    The intent of this Callable is to provide a hook to KGX applications wanting the
    namesake function of passively monitoring the graph data stream. As such, the Callable
    could simply tally up the number of times it is called with a NODE or an EDGE, then
    provide a suitable (quick!) report of that count back to the KGX application. The
    Callable (function/callable class) should not modify the record and should be of low
    complexity, so as not to introduce a large computational overhead to validation!
    """

    error_log = stderr

    def __init__(
        self,
        name="",
        node_facet_properties: Optional[List] = None,
        edge_facet_properties: Optional[List] = None,
        progress_monitor: Optional[Callable[[GraphEntityType, List], None]] = None,
        error_log: str = None,
        **kwargs,
    ):
        """
        GraphSummary constructor.

        Parameters
        ----------
        name: str
            (Graph) name assigned to the summary.
        node_facet_properties: Optional[List]
                A list of properties to facet on. For example, ``['provided_by']``
        edge_facet_properties: Optional[List]
                A list of properties to facet on. For example, ``['knowledge_source']``
        progress_monitor: Optional[Callable[[GraphEntityType, List], None]]
            Function given a peek at the current record being stream processed by the class wrapped Callable.
        error_log: str
            Where to write any graph processing error message (stderr, by default)

        """
        # formal arguments
        self.name = name

        self.nodes_processed = False

        self.node_stats: Dict = {
            TOTAL_NODES: 0,
            NODE_CATEGORIES: set(),
            NODE_ID_PREFIXES: set(),
            NODE_ID_PREFIXES_BY_CATEGORY: dict(),
            COUNT_BY_CATEGORY: dict(),
            COUNT_BY_ID_PREFIXES_BY_CATEGORY: dict(),
            COUNT_BY_ID_PREFIXES: dict(),
        }

        self.edges_processed: bool = False

        self.edge_stats: Dict = {
            TOTAL_EDGES: 0,
            EDGE_PREDICATES: set(),
            COUNT_BY_EDGE_PREDICATES: {"unknown": {"count": 0}},
            COUNT_BY_SPO: {},
        }

        self.node_facet_properties: Optional[List] = node_facet_properties
        if self.node_facet_properties:
            for facet_property in self.node_facet_properties:
                self.add_node_stat(facet_property, set())

        self.edge_facet_properties: Optional[List] = edge_facet_properties
        if self.edge_facet_properties:
            for facet_property in self.edge_facet_properties:
                self.edge_stats[facet_property] = set()

        self.progress_monitor: Optional[
            Callable[[GraphEntityType, List], None]
        ] = progress_monitor

        if error_log:
            GraphSummary.error_log = open(error_log, mode="w")

        # internal attributes
        self.node_catalog: Dict[str, List[int]] = dict()

        self.node_categories: Dict[str, GraphSummary.Category] = dict()

        # indexed internally with category index id '0'
        self.node_categories["unknown"] = GraphSummary.Category("unknown", self)

        self.graph_stats: Dict[str, Dict] = dict()

    def get_name(self):
        """
        Returns
        -------
        str
            Currently assigned knowledge graph name.
        """
        return self.name

    def __call__(self, entity_type: GraphEntityType, rec: List):
        """
        Transformer 'inspector' Callable, for analysing a stream of graph data.

        Parameters
        ----------
        entity_type: GraphEntityType
            indicates what kind of record being passed to the function for analysis.
        rec: Dict
            Complete data dictionary of the given record.

        """
        if self.progress_monitor:
            self.progress_monitor(entity_type, rec)
        if entity_type == GraphEntityType.EDGE:
            self.analyse_edge(*rec)
        elif entity_type == GraphEntityType.NODE:
            self.analyse_node(*rec)
        else:
            raise RuntimeError("Unexpected GraphEntityType: " + str(entity_type))

    class Category:
        """
        Internal class for compiling statistics about a distinct category.
        """

        # The 'category map' just associates a unique int catalog
        # index ('cid') value as a proxy for the full curie string,
        # to reduce storage in the main node catalog
        _category_curie_map: List[str] = list()

        def __init__(self, category_curie: str, summary):

            """
            GraphSummary.Category constructor.

            category: str
                Biolink Model category curie identifier.

            """
            if not (
                _category_curie_regexp.fullmatch(category_curie)
                or category_curie == "unknown"
            ):
                raise RuntimeError("Invalid Biolink category CURIE: " + category_curie)

            # generally, a Biolink  category class CURIE but also 'unknown'
            self.category_curie = category_curie

            # it is useful to point to the GraphSummary within
            # which this Category metadata is bring tracked...
            self.summary = summary

            # ...so that Category related entries at that
            # higher level may be properly initialized
            # for subsequent facet metadata access
            if not self.category_curie == "unknown":
                self.summary.node_stats[NODE_CATEGORIES].add(self.category_curie)
            self.summary.node_stats[NODE_ID_PREFIXES_BY_CATEGORY][
                self.category_curie
            ] = set()
            self.summary.node_stats[COUNT_BY_CATEGORY][self.category_curie] = {
                "count": 0
            }

            if self.category_curie not in self._category_curie_map:
                self._category_curie_map.append(self.category_curie)
            self.category_stats: Dict[str, Any] = dict()
            self.category_stats["count"]: int = 0
            self.category_stats["count_by_source"]: Dict[str, int] = {"unknown": 0}
            self.category_stats["count_by_id_prefix"]: Dict[str, int] = dict()

        def get_name(self) -> str:
            """
            Returns
            -------
            str
                Biolink CURIE name of the category.
            """
            return self.category_curie

        def get_cid(self) -> int:
            """
            Returns
            -------
            int
                Internal GraphSummary index id for tracking a Category.
            """
            return self._category_curie_map.index(self.category_curie)

        @classmethod
        def _get_category_curie_by_index(cls, cid: int) -> str:
            """
            Parameters
            ----------
            cid: int
                Internal GraphSummary index id for tracking a Category.

            Returns
            -------
            str
                Curie identifier of the Category.
            """
            return cls._category_curie_map[cid]

        def get_id_prefixes(self) -> Set:
            """
            Returns
            -------
            Set[str]
                Set of identifier prefix (strings) used by nodes of this Category.
            """
            return set(self.category_stats["count_by_id_prefix"].keys())

        def get_count_by_id_prefixes(self):
            """
            Returns
            -------
            int
                Count of nodes by id_prefixes for nodes which have this category.
            """
            return self.category_stats["count_by_id_prefix"]

        def get_count(self):
            """
            Returns
            -------
            int
                Count of nodes which have this category.
            """
            return self.category_stats["count"]

        def _capture_prefix(self, n: str):
            prefix = PrefixManager.get_prefix(n)
            if not prefix:
                print(
                    f"Warning: node id {n} has no CURIE prefix",
                    file=GraphSummary.error_log,
                )
            else:
                if prefix in self.category_stats["count_by_id_prefix"]:
                    self.category_stats["count_by_id_prefix"][prefix] += 1
                else:
                    self.category_stats["count_by_id_prefix"][prefix] = 1

        def _capture_knowledge_source(self, data: Dict):
            if "provided_by" in data:
                for s in data["provided_by"]:
                    if s in self.category_stats["count_by_source"]:
                        self.category_stats["count_by_source"][s] += 1
                    else:
                        self.category_stats["count_by_source"][s] = 1
            else:
                self.category_stats["count_by_source"]["unknown"] += 1

        def analyse_node_category(self, summary, n, data):
            """
            Analyse metadata of a given graph node record of this category.

            Parameters
            ----------
            summary: GraphSummary
                GraphSunmmary within which the Category is being analysed.
            n: str
                Curie identifier of the node record (not used here).
            data: Dict
                Complete data dictionary of node record fields.

            """
            self.category_stats["count"] += 1

            self._capture_prefix(n)

            self._capture_knowledge_source(data)

            if summary.node_facet_properties:
                for facet_property in summary.node_facet_properties:
                    summary.node_stats = summary.get_facet_counts(
                        data,
                        summary.node_stats,
                        COUNT_BY_CATEGORY,
                        self.category_curie,
                        facet_property,
                    )

        def json_object(self):
            """
            Returns
            -------
            Dict[str, Any]
                Returns JSON friendly metadata for this category.,
            """
            return {
                "id_prefixes": list(self.category_stats["count_by_id_prefix"].keys()),
                "count": self.category_stats["count"],
                "count_by_source": self.category_stats["count_by_source"],
                "count_by_id_prefix": self.category_stats["count_by_id_prefix"],
            }

    def get_category(self, category_curie: str) -> Category:
        """
        Counts the number of distinct (Biolink) categories encountered
        in the knowledge graph (not including those of 'unknown' category)

        Parameters
        ----------
        category_curie: str
            Curie identifier for the (Biolink) category.

        Returns
        -------
        Category
            MetaKnowledgeGraph.Category object for a given Biolink category.
        """
        return self.node_stats[category_curie]

    def _process_category_field(self, category_field: str, n: str, data: Dict):

        # we note here that category_curie *may be*
        # a piped '|' set of Biolink category CURIE values
        category_list = category_field.split("|")

        # analyse them each independently...
        for category_curie in category_list:

            if category_curie not in self.node_categories:
                try:
                    self.node_categories[category_curie] = self.Category(
                        category_curie, self
                    )
                except RuntimeError:
                    _parse_warning("Invalid category CURIE", category_curie)
                    continue

            category_record = self.node_categories[category_curie]
            category_idx: int = category_record.get_cid()
            if category_idx not in self.node_catalog[n]:
                self.node_catalog[n].append(category_idx)
            category_record.analyse_node_category(self, n, data)

        #
        # Moved this computation from the 'analyse_node_category() method above
        #
        # if self.node_facet_properties:
        #     for facet_property in self.node_facet_properties:
        #         self.node_stats = self.get_facet_counts(
        #             data, self.node_stats, COUNT_BY_CATEGORY, category_curie, facet_property
        #         )

    def analyse_node(self, n, data):
        """
        Analyse metadata of one graph node record.

        Parameters
        ----------
        n: str
            Curie identifier of the node record (not used here).
        data: Dict
            Complete data dictionary of node record fields.

        """
        if n in self.node_catalog:
            # Report duplications of node records, as discerned from node id.
            _parse_warning(
                "Duplicate node identifier", n, "encountered in input node data"
            )
            return
        else:
            self.node_catalog[n] = list()

        if "category" in data and data["category"]:
            categories = data["category"]

        else:
            categories = ["unknown"]
            print(
                "Node with identifier '"
                + n
                + "' is missing its 'category' value? Tagging it as 'unknown'",
                file=GraphSummary.error_log,
            )

        # analyse them each independently...
        for category_field in categories:
            self._process_category_field(category_field, n, data)

    def _capture_predicate(self, data: Dict) -> Optional[str]:
        if "predicate" not in data:
            self.edge_stats[COUNT_BY_EDGE_PREDICATES]["unknown"]["count"] += 1
            predicate = "unknown"
        else:
            predicate = data["predicate"]

            if not _predicate_curie_regexp.fullmatch(predicate):
                _parse_warning("Invalid  predicate CURIE", predicate)
                return None

            self.edge_stats[EDGE_PREDICATES].add(predicate)
            if predicate in self.edge_stats[COUNT_BY_EDGE_PREDICATES]:
                self.edge_stats[COUNT_BY_EDGE_PREDICATES][predicate]["count"] += 1
            else:
                self.edge_stats[COUNT_BY_EDGE_PREDICATES][predicate] = {"count": 1}

            if self.edge_facet_properties:
                for facet_property in self.edge_facet_properties:
                    self.edge_stats = self.get_facet_counts(
                        data,
                        self.edge_stats,
                        COUNT_BY_EDGE_PREDICATES,
                        predicate,
                        facet_property,
                    )

        return predicate

    def _process_triple(
        self, subject_category: str, predicate: str, object_category: str, data: Dict
    ):
        # Process the 'valid' S-P-O triple here...
        key = f"{subject_category}-{predicate}-{object_category}"
        if key in self.edge_stats[COUNT_BY_SPO]:
            self.edge_stats[COUNT_BY_SPO][key]["count"] += 1
        else:
            self.edge_stats[COUNT_BY_SPO][key] = {"count": 1}

        if self.edge_facet_properties:
            for facet_property in self.edge_facet_properties:
                self.edge_stats = self.get_facet_counts(
                    data, self.edge_stats, COUNT_BY_SPO, key, facet_property
                )

    def analyse_edge(self, u: str, v: str, k: str, data: Dict):
        """
        Analyse metadata of one graph edge record.

        Parameters
        ----------
        u: str
            Subject node curie identifier of the edge.
        v: str
            Subject node curie identifier of the edge.
        k: str
            Key identifier of the edge record (not used here).
        data: Dict
            Complete data dictionary of edge record fields.

        """
        # we blissfully now assume that all the nodes of a
        # graph stream were analysed first by the GraphSummary
        # before the edges are analysed, thus we can test for
        # node 'n' existence internally, by identifier.

        self.edge_stats[TOTAL_EDGES] += 1

        predicate: str = self._capture_predicate(data)

        if u not in self.node_catalog:
            _parse_warning("Edge 'subject' node ID", u, "not found in node catalog")
            # removing from edge count
            self.edge_stats[TOTAL_EDGES] -= 1
            self.edge_stats[COUNT_BY_EDGE_PREDICATES]["unknown"]["count"] -= 1
            return

        for subj_cat_idx in self.node_catalog[u]:

            subject_category = self.Category._get_category_curie_by_index(subj_cat_idx)

            if v not in self.node_catalog:
                _parse_warning("Edge 'object' node ID", v, "not found in node catalog")
                self.edge_stats[TOTAL_EDGES] -= 1
                self.edge_stats[COUNT_BY_EDGE_PREDICATES]["unknown"]["count"] -= 1
                return

            for obj_cat_idx in self.node_catalog[v]:

                object_category = self.Category._get_category_curie_by_index(
                    obj_cat_idx
                )

                self._process_triple(subject_category, predicate, object_category, data)

    def _compile_prefix_stats_by_category(self, category_curie: str):
        for prefix in self.node_stats[COUNT_BY_ID_PREFIXES_BY_CATEGORY][category_curie]:
            if prefix not in self.node_stats[COUNT_BY_ID_PREFIXES]:
                self.node_stats[COUNT_BY_ID_PREFIXES][prefix] = 0
            self.node_stats[COUNT_BY_ID_PREFIXES][prefix] += self.node_stats[
                COUNT_BY_ID_PREFIXES_BY_CATEGORY
            ][category_curie][prefix]

    def _compile_category_stats(self, node_category: Category):
        category_curie = node_category.get_name()

        self.node_stats[COUNT_BY_CATEGORY][category_curie][
            "count"
        ] = node_category.get_count()

        id_prefixes = node_category.get_id_prefixes()
        self.node_stats[NODE_ID_PREFIXES_BY_CATEGORY][category_curie] = id_prefixes
        self.node_stats[NODE_ID_PREFIXES].update(id_prefixes)

        self.node_stats[COUNT_BY_ID_PREFIXES_BY_CATEGORY][
            category_curie
        ] = node_category.get_count_by_id_prefixes()

        self._compile_prefix_stats_by_category(category_curie)

    def get_node_stats(self) -> Dict[str, Any]:
        """
        Returns
        -------
        Dict[str, Any]
            Statistics for the nodes in the graph.
        """
        if not self.nodes_processed:

            self.nodes_processed = True

            for node_category in self.node_categories.values():
                self._compile_category_stats(node_category)

            self.node_stats[NODE_CATEGORIES] = sorted(
                list(self.node_stats[NODE_CATEGORIES])
            )

            if self.node_facet_properties:
                for facet_property in self.node_facet_properties:
                    self.node_stats[facet_property] = sorted(
                        list(self.node_stats[facet_property])
                    )

            if not self.node_stats[TOTAL_NODES]:
                self.node_stats[TOTAL_NODES] = len(self.node_catalog)

        return self.node_stats

    def add_node_stat(self, tag: str, value: Any):
        """
        Compile/add a nodes statistic for a given tag = value annotation of the node.

        :param tag:
        :param value:
        :return:

        Parameters
        ----------
        tag: str
            Tag label for the annotation.
        value: Any
             Value of the specific tag annotation.

        """
        self.node_stats[tag] = value

    def get_edge_stats(self) -> Dict[str, Any]:
        # Not sure if this is "safe" but assume that edge_stats may be finalized
        # and cached once after the first time the edge stats are accessed
        if not self.edges_processed:
            self.edges_processed = True

            self.edge_stats[EDGE_PREDICATES] = sorted(
                list(self.edge_stats[EDGE_PREDICATES])
            )

            if self.edge_facet_properties:
                for facet_property in self.edge_facet_properties:
                    self.edge_stats[facet_property] = sorted(
                        list(self.edge_stats[facet_property])
                    )

        return self.edge_stats

    def _wrap_graph_stats(
        self,
        graph_name: str,
        node_stats: Dict[str, Any],
        edge_stats: Dict[str, Any],
    ):
        # Utility wrapper function to support DRY code below.
        if not self.graph_stats:
            self.graph_stats = {
                "graph_name": graph_name,
                "node_stats": node_stats,
                "edge_stats": edge_stats,
            }
        return self.graph_stats

    def get_graph_summary(self, name: str = None, **kwargs) -> Dict:
        """
        Similar to summarize_graph except that the node and edge statistics are already captured
        in the GraphSummary class instance (perhaps by Transformer.process() stream inspection)
        and therefore, the data structure simply needs to be 'finalized' for saving or similar use.

        Parameters
        ----------
        name: Optional[str]
            Name for the graph (if being renamed)
        kwargs: Dict
            Any additional arguments (ignored in this method at present)

        Returns
        -------
        Dict
            A knowledge map dictionary corresponding to the graph

        """
        return self._wrap_graph_stats(
            graph_name=name if name else self.name,
            node_stats=self.get_node_stats(),
            edge_stats=self.get_edge_stats(),
        )

    def summarize_graph(self, graph: BaseGraph) -> Dict:
        """
        Summarize the entire graph.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph

        Returns
        -------
        Dict
            The stats dictionary

        """
        return self._wrap_graph_stats(
            graph_name=self.name if self.name else graph.name,
            node_stats=self.summarize_graph_nodes(graph),
            edge_stats=self.summarize_graph_edges(graph),
        )

    def summarize_graph_nodes(self, graph: BaseGraph) -> Dict:
        """
        Summarize the nodes in a graph.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph

        Returns
        -------
        Dict
            The node stats
        """
        for n, data in graph.nodes(data=True):
            self.analyse_node(n, data)

        return self.get_node_stats()

    def summarize_graph_edges(self, graph: BaseGraph) -> Dict:
        """
        Summarize the edges in a graph.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph

        Returns
        -------
        Dict
            The edge stats
        """
        for u, v, k, data in graph.edges(keys=True, data=True):
            self.analyse_edge(u, v, k, data)

        return self.get_edge_stats()

    def _compile_facet_stats(
        self, stats: Dict, x: str, y: str, facet_property: str, value: str
    ):

        if facet_property not in stats[x][y]:
            stats[x][y][facet_property] = {}

        if value in stats[x][y][facet_property]:
            stats[x][y][facet_property][value]["count"] += 1
        else:
            stats[x][y][facet_property][value] = {"count": 1}
            stats[facet_property].update([value])

    def get_facet_counts(
        self, data: Dict, stats: Dict, x: str, y: str, facet_property: str
    ) -> Dict:
        """
        Facet on ``facet_property`` and record the count for ``stats[x][y][facet_property]``.

        Parameters
        ----------
        data: dict
            Node/edge data dictionary
        stats: dict
            The stats dictionary
        x: str
            first key
        y: str
            second key
        facet_property: str
            The property to facet on

        Returns
        -------
        Dict
            The stats dictionary

        """
        if facet_property in data:
            if isinstance(data[facet_property], list):
                for k in data[facet_property]:
                    self._compile_facet_stats(stats, x, y, facet_property, k)
            else:
                k = data[facet_property]
                self._compile_facet_stats(stats, x, y, facet_property, k)
        else:
            self._compile_facet_stats(stats, x, y, facet_property, "unknown")
        return stats

    def save(self, file, name: str = None, file_format: str = "yaml"):
        """
        Save the current GraphSummary to a specified (open) file (device).

        Parameters
        ----------
        file: File
            Text file handler open for writing.
        name: str
            Optional string to which to (re-)name the graph.
        file_format:  str
            Text output format ('json' or 'yaml') for the saved meta knowledge graph  (default: 'json')

        Returns
        -------
        None
        """
        stats = self.get_graph_summary(name)
        if not file_format or file_format == "yaml":
            yaml.dump(stats, file)
        else:
            dump(stats, file, indent=4, default=gs_default)


def generate_graph_stats(
    graph: BaseGraph,
    graph_name: str,
    filename: str,
    node_facet_properties: Optional[List] = None,
    edge_facet_properties: Optional[List] = None,
) -> None:
    """
    Generate stats from Graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    graph_name: str
        Name for the graph
    filename: str
        Filename to write the stats to
    node_facet_properties: Optional[List]
        A list of properties to facet on. For example, ``['provided_by']``
    edge_facet_properties: Optional[List]
        A list of properties to facet on. For example, ``['knowledge_source']``

    """
    stats = summarize_graph(
        graph, graph_name, node_facet_properties, edge_facet_properties
    )
    with open(filename, "w") as gsh:
        yaml.dump(stats, gsh)


def summarize_graph(
    graph: BaseGraph,
    name: str = None,
    node_facet_properties: Optional[List] = None,
    edge_facet_properties: Optional[List] = None,
) -> Dict:
    """
    Summarize the entire graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    name: str
        Name for the graph
    node_facet_properties: Optional[List]
        A list of properties to facet on. For example, ``['provided_by']``
    edge_facet_properties: Optional[List]
        A list of properties to facet on. For example, ``['knowledge_source']``

    Returns
    -------
    Dict
        The stats dictionary

    """
    gs = GraphSummary(name, node_facet_properties, edge_facet_properties)
    return gs.summarize_graph(graph)



from typing import Dict, Any, Optional, List, Generator

from kgx.graph.base_graph import BaseGraph
from networkx import (
    MultiDiGraph,
    set_node_attributes,
    relabel_nodes,
    set_edge_attributes,
    get_node_attributes,
    get_edge_attributes,
)

from kgx.utils.kgx_utils import prepare_data_dict


class NxGraph(BaseGraph):
    """
    NxGraph is a wrapper that provides methods to interact with a networkx.MultiDiGraph.

    NxGraph extends kgx.graph.base_graph.BaseGraph and implements all the methods from BaseGraph.
    """

    def __init__(self):
        super().__init__()
        self.graph = MultiDiGraph()
        self.name = None

    def add_node(self, node: str, **kwargs: Any) -> None:
        """
        Add a node to the graph.

        Parameters
        ----------
        node: str
            Node identifier
        **kwargs: Any
            Any additional node properties

        """
        if "data" in kwargs:
            data = kwargs["data"]
        else:
            data = kwargs
        self.graph.add_node(node, **data)

    def add_edge(
        self, subject_node: str, object_node: str, edge_key: str = None, **kwargs: Any
    ) -> None:
        """
        Add an edge to the graph.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key
        kwargs: Any
            Any additional edge properties

        """
        if "data" in kwargs:
            data = kwargs["data"]
        else:
            data = kwargs
        return self.graph.add_edge(subject_node, object_node, key=edge_key, **data)

    def add_node_attribute(self, node: str, attr_key: str, attr_value: Any) -> None:
        """
        Add an attribute to a given node.

        Parameters
        ----------
        node: str
            The node identifier
        attr_key: str
            The key for an attribute
        attr_value: Any
            The value corresponding to the key

        """
        self.graph.add_node(node, **{attr_key: attr_value})

    def add_edge_attribute(
        self,
        subject_node: str,
        object_node: str,
        edge_key: Optional[str],
        attr_key: str,
        attr_value: Any,
    ) -> None:
        """
        Add an attribute to a given edge.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key
        attr_key: str
            The attribute key
        attr_value: Any
            The attribute value

        """
        self.graph.add_edge(
            subject_node, object_node, key=edge_key, **{attr_key: attr_value}
        )

    def update_node_attribute(
        self, node: str, attr_key: str, attr_value: Any, preserve: bool = False
    ) -> Dict:
        """
        Update an attribute of a given node.

        Parameters
        ----------
        node: str
            The node identifier
        attr_key: str
            The key for an attribute
        attr_value: Any
            The value corresponding to the key
        preserve: bool
            Whether or not to preserve existing values for the given attr_key

        Returns
        -------
        Dict
            A dictionary corresponding to the updated node properties

        """
        node_data = self.graph.nodes[node]
        updated = prepare_data_dict(
            node_data, {attr_key: attr_value}, preserve=preserve
        )
        self.graph.add_node(node, **updated)
        return updated

    def update_edge_attribute(
        self,
        subject_node: str,
        object_node: str,
        edge_key: Optional[str],
        attr_key: str,
        attr_value: Any,
        preserve: bool = False,
    ) -> Dict:
        """
        Update an attribute of a given edge.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key
        attr_key: str
            The attribute key
        attr_value: Any
            The attribute value
        preserve: bool
            Whether or not to preserve existing values for the given attr_key

        Returns
        -------
        Dict
            A dictionary corresponding to the updated edge properties

        """
        e = self.graph.edges(
            (subject_node, object_node, edge_key), keys=True, data=True
        )
        edge_data = list(e)[0][3]
        updated = prepare_data_dict(edge_data, {attr_key: attr_value}, preserve)
        self.graph.add_edge(subject_node, object_node, key=edge_key, **updated)
        return updated

    def get_node(self, node: str) -> Dict:
        """
        Get a node and its properties.

        Parameters
        ----------
        node: str
            The node identifier

        Returns
        -------
        Dict
            The node dictionary

        """
        n = {}
        if self.graph.has_node(node):
            n = self.graph.nodes[node]
        return n

    def get_edge(
        self, subject_node: str, object_node: str, edge_key: Optional[str] = None
    ) -> Dict:
        """
        Get an edge and its properties.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key

        Returns
        -------
        Dict
            The edge dictionary

        """
        e = {}
        if self.graph.has_edge(subject_node, object_node, edge_key):
            e = self.graph.get_edge_data(subject_node, object_node, edge_key)
        return e

    def nodes(self, data: bool = True) -> Dict:
        """
        Get all nodes in a graph.

        Parameters
        ----------
        data: bool
            Whether or not to fetch node properties

        Returns
        -------
        Dict
            A dictionary of nodes

        """
        return self.graph.nodes(data)

    def edges(self, keys: bool = False, data: bool = True) -> Dict:
        """
        Get all edges in a graph.

        Parameters
        ----------
        keys: bool
            Whether or not to include edge keys
        data: bool
            Whether or not to fetch node properties

        Returns
        -------
        Dict
            A dictionary of edges

        """
        return self.graph.edges(keys=keys, data=data)

    def in_edges(self, node: str, keys: bool = False, data: bool = False) -> List:
        """
        Get all incoming edges for a given node.

        Parameters
        ----------
        node: str
            The node identifier
        keys: bool
            Whether or not to include edge keys
        data: bool
            Whether or not to fetch node properties

        Returns
        -------
        List
            A list of edges

        """
        return self.graph.in_edges(node, keys=keys, data=data)

    def out_edges(self, node: str, keys: bool = False, data: bool = False) -> List:
        """
        Get all outgoing edges for a given node.

        Parameters
        ----------
        node: str
            The node identifier
        keys: bool
            Whether or not to include edge keys
        data: bool
            Whether or not to fetch node properties

        Returns
        -------
        List
            A list of edges

        """
        return self.graph.out_edges(node, keys=keys, data=data)

    def nodes_iter(self) -> Generator:
        """
        Get an iterable to traverse through all the nodes in a graph.

        Returns
        -------
        Generator
            A generator for nodes where each element is a Tuple that
            contains (node_id, node_data)

        """
        for n in self.graph.nodes(data=True):
            yield n

    def edges_iter(self) -> Generator:
        """
        Get an iterable to traverse through all the edges in a graph.

        Returns
        -------
        Generator
            A generator for edges where each element is a 4-tuple that
            contains (subject, object, edge_key, edge_data)

        """
        for u, v, k, data in self.graph.edges(keys=True, data=True):
            yield u, v, k, data

    def remove_node(self, node: str) -> None:
        """
        Remove a given node from the graph.

        Parameters
        ----------
        node: str
            The node identifier

        """
        self.graph.remove_node(node)

    def remove_edge(
        self, subject_node: str, object_node: str, edge_key: Optional[str] = None
    ) -> None:
        """
        Remove a given edge from the graph.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key

        """
        self.graph.remove_edge(subject_node, object_node, edge_key)

    def has_node(self, node: str) -> bool:
        """
        Check whether a given node exists in the graph.

        Parameters
        ----------
        node: str
            The node identifier

        Returns
        -------
        bool
            Whether or not the given node exists

        """
        return self.graph.has_node(node)

    def has_edge(
        self, subject_node: str, object_node: str, edge_key: Optional[str] = None
    ) -> bool:
        """
        Check whether a given edge exists in the graph.

        Parameters
        ----------
        subject_node: str
            The subject (source) node
        object_node: str
            The object (target) node
        edge_key: Optional[str]
            The edge key

        Returns
        -------
        bool
            Whether or not the given edge exists

        """
        return self.graph.has_edge(subject_node, object_node, key=edge_key)

    def number_of_nodes(self) -> int:
        """
        Returns the number of nodes in a graph.

        Returns
        -------
        int

        """
        return self.graph.number_of_nodes()

    def number_of_edges(self) -> int:
        """
        Returns the number of edges in a graph.

        Returns
        -------
        int

        """
        return self.graph.number_of_edges()

    def degree(self):
        """
        Get the degree of all the nodes in a graph.
        """
        return self.graph.degree()

    def clear(self) -> None:
        """
        Remove all the nodes and edges in the graph.
        """
        self.graph.clear()

    @staticmethod
    def set_node_attributes(graph: BaseGraph, attributes: Dict) -> None:
        """
        Set nodes attributes from a dictionary of key-values.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to modify
        attributes: Dict
            A dictionary of node identifier to key-value pairs

        """
        return set_node_attributes(graph.graph, attributes)

    @staticmethod
    def set_edge_attributes(graph: BaseGraph, attributes: Dict) -> None:
        """
        Set nodes attributes from a dictionary of key-values.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to modify
        attributes: Dict
            A dictionary of node identifier to key-value pairs

        Returns
        -------
        Any

        """
        return set_edge_attributes(graph.graph, attributes)

    @staticmethod
    def get_node_attributes(graph: BaseGraph, attr_key: str) -> Dict:
        """
        Get all nodes that have a value for the given attribute ``attr_key``.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to modify
        attr_key: str
            The attribute key

        Returns
        -------
        Dict
            A dictionary where nodes are the keys and the values
            are the attribute values for ``key``

        """
        return get_node_attributes(graph.graph, attr_key)

    @staticmethod
    def get_edge_attributes(graph: BaseGraph, attr_key: str) -> Dict:
        """
        Get all edges that have a value for the given attribute ``attr_key``.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to modify
        attr_key: str
            The attribute key

        Returns
        -------
        Dict
            A dictionary where edges are the keys and the values
            are the attribute values for ``attr_key``

        """
        return get_edge_attributes(graph.graph, attr_key)

    @staticmethod
    def relabel_nodes(graph: BaseGraph, mapping: Dict) -> None:
        """
        Relabel identifiers for a series of nodes based on mappings.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to modify
        mapping: Dict
            A dictionary of mapping where the key is the old identifier
            and the value is the new identifier.

        """
        relabel_nodes(graph.graph, mapping, copy=False)



import itertools
from typing import Any, Dict, List, Optional, Iterator, Tuple, Generator

from neo4jrestclient.client import Node, Relationship, GraphDatabase
from neo4jrestclient.query import CypherException

from kgx.config import get_logger
from kgx.source.source import Source
from kgx.utils.kgx_utils import (
    generate_uuid,
    generate_edge_key,
    validate_node,
    validate_edge,
    sanitize_import,
    knowledge_provenance_properties,
)

log = get_logger()


class NeoSource(Source):
    """
    NeoSource is responsible for reading data as records
    from a Neo4j instance.
    """

    def __init__(self):
        super().__init__()
        self.http_driver = None
        self.node_count = 0
        self.edge_count = 0
        self.seen_nodes = set()

    def parse(
        self,
        uri: str,
        username: str,
        password: str,
        node_filters: Dict = None,
        edge_filters: Dict = None,
        start: int = 0,
        end: int = None,
        is_directed: bool = True,
        page_size: int = 50000,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from Neo4j instance and yields records

        Parameters
        ----------
        uri: str
            The URI for the Neo4j instance.
            For example, http://localhost:7474
        username: str
            The username
        password: str
            The password
        node_filters: Dict
            Node filters
        edge_filters: Dict
            Edge filters
        start: int
            Number of records to skip before streaming
        end: int
            Total number of records to fetch
        is_directed: bool
            Whether or not the edges should be treated as directed
        page_size: int
            The size of each page/batch fetched from Neo4j (``50000``)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for records

        """
        self.http_driver: GraphDatabase = GraphDatabase(
            uri, username=username, password=password
        )

        self.set_provenance_map(kwargs)

        kwargs["is_directed"] = is_directed
        self.node_filters = node_filters
        self.edge_filters = edge_filters
        for page in self.get_pages(
            self.get_nodes, start, end, page_size=page_size, **kwargs
        ):
            yield from self.load_nodes(page)
        for page in self.get_pages(
            self.get_edges, start, end, page_size=page_size, **kwargs
        ):
            yield from self.load_edges(page)

    def count(self, is_directed: bool = True) -> int:
        """
        Get the total count of records to be fetched from the Neo4j database.

        Parameters
        ----------
        is_directed: bool
            Are edges directed or undirected.
            ``True``, by default, since edges in most cases are directed.

        Returns
        -------
        int
            The total count of records

        """
        direction = "->" if is_directed else "-"
        query = f"MATCH (s)-[p]{direction}(o)"

        if self.edge_filters:
            qs = []
            if "subject_category" in self.edge_filters:
                qs.append(
                    f"({self.format_edge_filter(self.edge_filters, 'subject_category', 's', ':', 'OR')})"
                )
            if "object_category" in self.edge_filters:
                qs.append(
                    f"({self.format_edge_filter(self.edge_filters, 'object_category', 'o', ':', 'OR')})"
                )
            if "predicate" in self.edge_filters:
                qs.append(
                    f"({self.format_edge_filter(self.edge_filters, 'predicate', 'p', '.')})"
                )
            for ksf in knowledge_provenance_properties:
                if ksf in self.edge_filters:
                    qs.append(
                        f"({self.format_edge_filter(self.edge_filters, ksf, 'p', '.', 'OR')})"
                    )
            query = " WHERE "
            query += " AND ".join(qs)
        query += f" RETURN COUNT(*) AS count"
        log.debug(query)
        query_result: Any
        counts: int = 0
        try:
            query_result = self.http_driver.query(query)
            for result in query_result:
                counts = result[0]
        except CypherException as ce:
            log.error(ce)
        return counts

    def get_nodes(self, skip: int = 0, limit: int = 0, **kwargs: Any) -> List:
        """
        Get a page of nodes from the Neo4j database.

        Parameters
        ----------
        skip: int
            Records to skip
        limit: int
            Total number of records to query for
        kwargs: Any
            Any additional arguments

        Returns
        -------
        List
            A list of nodes

        """
        query = f"MATCH (n)"

        if self.node_filters:
            qs = []
            if "category" in self.node_filters:
                qs.append(
                    f"({self.format_node_filter(self.node_filters, 'category', 'n', ':', 'OR')})"
                )
            if "provided_by" in self.node_filters:
                qs.append(
                    f"({self.format_node_filter(self.node_filters, 'provided_by', 'n', '.', 'OR')})"
                )
            query += " WHERE "
            query += " AND ".join(qs)

        query += f" RETURN n SKIP {skip}"

        if limit:
            query += f" LIMIT {limit}"

        log.debug(query)
        nodes = []
        try:
            results = self.http_driver.query(query, returns=Node, data_contents=True)
            if results:
                nodes = [node[0] for node in results.rows]
        except CypherException as ce:
            log.error(ce)
        return nodes

    def get_edges(
        self, skip: int = 0, limit: int = 0, is_directed: bool = True, **kwargs: Any
    ) -> List:
        """
        Get a page of edges from the Neo4j database.

        Parameters
        ----------
        skip: int
            Records to skip
        limit: int
            Total number of records to query for
        is_directed: bool
            Are edges directed or undirected (``True``, by default, since edges in most cases are directed)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        List
            A list of 3-tuples

        """
        direction = "->" if is_directed else "-"
        query = f"MATCH (s)-[p]{direction}(o)"

        if self.edge_filters:
            qs = []
            if "subject_category" in self.edge_filters:
                qs.append(
                    f"({self.format_edge_filter(self.edge_filters, 'subject_category', 's', ':', 'OR')})"
                )
            if "object_category" in self.edge_filters:
                qs.append(
                    f"({self.format_edge_filter(self.edge_filters, 'object_category', 'o', ':', 'OR')})"
                )
            if "predicate" in self.edge_filters:
                qs.append(
                    f"({self.format_edge_filter(self.edge_filters, 'predicate', 'p', '.')})"
                )
            for ksf in knowledge_provenance_properties:
                if ksf in self.edge_filters:
                    qs.append(
                        f"({self.format_edge_filter(self.edge_filters, ksf, 'p', '.', 'OR')})"
                    )
            query += " WHERE "
            query += " AND ".join(qs)
        query += f" RETURN s, p, o SKIP {skip}"

        if limit:
            query += f" LIMIT {limit}"

        log.debug(query)
        edges = []
        try:
            results = self.http_driver.query(
                query, returns=(Node, Relationship, Node), data_contents=True
            )
            if results:
                edges = [x for x in results.rows]
        except CypherException as ce:
            log.error(ce)

        return edges

    def load_nodes(self, nodes: List) -> None:
        """
        Load nodes into an instance of BaseGraph

        Parameters
        ----------
        nodes: List
            A list of nodes

        """
        for node in nodes:
            if node["id"] not in self.seen_nodes:
                yield self.load_node(node)

    def load_node(self, node: Dict) -> Tuple:
        """
        Load node into an instance of BaseGraph

        Parameters
        ----------
        node: Dict
            A node

        Returns
        -------
        Tuple
            A tuple with node ID and node data

        """
        self.node_count += 1
        # TODO: remove the seen_nodes
        self.seen_nodes.add(node["id"])

        self.set_node_provenance(node)

        node = validate_node(node)
        node = sanitize_import(node.copy())
        self.node_properties.update(node.keys())
        return node["id"], node

    def load_edges(self, edges: List) -> None:
        """
        Load edges into an instance of BaseGraph

        Parameters
        ----------
        edges: List
            A list of edge records

        """
        for record in edges:
            self.edge_count += 1
            subject_node = record[0]
            edge = record[1]
            object_node = record[2]

            if "subject" not in edge:
                edge["subject"] = subject_node["id"]
            if "object" not in edge:
                edge["object"] = object_node["id"]

            s = self.load_node(subject_node)
            o = self.load_node(object_node)
            objs = [s, o]
            objs.append(self.load_edge([s[1], edge, o[1]]))
            for o in objs:
                yield o

    def load_edge(self, edge_record: List) -> Tuple:
        """
        Load an edge into an instance of BaseGraph

        Parameters
        ----------
        edge_record: List
            A 4-tuple edge record

        Returns
        -------
        Tuple
            A tuple with subject ID, object ID, edge key, and edge data

        """

        subject_node = edge_record[0]
        edge = edge_record[1]
        object_node = edge_record[2]

        self.set_edge_provenance(edge)

        if "id" not in edge.keys():
            edge["id"] = generate_uuid()
        key = generate_edge_key(
            subject_node["id"], edge["predicate"], object_node["id"]
        )
        edge = validate_edge(edge)
        edge = sanitize_import(edge.copy())
        self.edge_properties.update(edge.keys())
        return subject_node["id"], object_node["id"], key, edge

    def get_pages(
        self,
        query_function,
        start: int = 0,
        end: Optional[int] = None,
        page_size: int = 50000,
        **kwargs: Any,
    ) -> Iterator:
        """
        Get pages of size ``page_size`` from Neo4j.
        Returns an iterator of pages where number of pages is (``end`` - ``start``)/``page_size``

        Parameters
        ----------
        query_function: func
            The function to use to fetch records. Usually this is ``self.get_nodes`` or ``self.get_edges``
        start: int
            Start for pagination
        end: Optional[int]
            End for pagination
        page_size: int
            Size of each page (``10000``, by default)
        kwargs: Dict
            Any additional arguments that might be relevant for ``query_function``

        Returns
        -------
        Iterator
            An iterator for a list of records from Neo4j. The size of the list is ``page_size``

        """
        # TODO: use async
        # itertools.count(0) starts counting from zero, and would run indefinitely without a return statement.
        # it's distinguished from applying a while loop via providing an index which is formative with the for statement
        for i in itertools.count(0):
            # First halt condition: page pointer exceeds the number of values allowed to be returned in total
            skip = start + (page_size * i)
            limit = page_size if end is None or skip + page_size <= end else end - skip
            if limit <= 0:
                return
            # execute query_function to get records
            records = query_function(skip=skip, limit=limit, **kwargs)
            # Second halt condition: no more data available
            if records:
                """
                * Yield halts execution until next call
                * Thus, the function continues execution upon next call
                * Therefore, a new page is calculated before record is instantiated again
                """
                yield records
            else:
                return

    @staticmethod
    def format_node_filter(
        node_filters: Dict,
        key: str,
        variable: Optional[str] = None,
        prefix: Optional[str] = None,
        op: Optional[str] = None,
    ) -> str:
        """
        Get the value for node filter as defined by ``key``.
        This is used as a convenience method for generating cypher queries.

        Parameters
        ----------
        node_filters: Dict
            All node filters
        key: str
            Name of the node filter
        variable: Optional[str]
            Variable binding for cypher query
        prefix: Optional[str]
            Prefix for the cypher
        op: Optional[str]
            The operator

        Returns
        -------
        str
            Value corresponding to the given node filter ``key``, formatted for CQL

        """
        value = ""
        if key in node_filters and node_filters[key]:
            if isinstance(node_filters[key], (list, set, tuple)):
                if key in {"category"}:
                    formatted = [f"{variable}{prefix}`{x}`" for x in node_filters[key]]
                    value = f" {op} ".join(formatted)
                elif key == "provided_by":
                    formatted = [
                        f"'{x}' IN {variable}{prefix}{'provided_by'}"
                        for x in node_filters["provided_by"]
                    ]
                    value = f" {op} ".join(formatted)
                else:
                    formatted = []
                    for v in node_filters[key]:
                        formatted.append(f"{variable}{prefix}{key} = '{v}'")
                    value = f" {op} ".join(formatted)
            elif isinstance(node_filters[key], str):
                value = f"{variable}{prefix}{key} = '{node_filters[key]}'"
            else:
                log.error(
                    f"Unexpected {key} node filter of type {type(node_filters[key])}"
                )
        return value

    @staticmethod
    def format_edge_filter(
        edge_filters: Dict,
        key: str,
        variable: Optional[str] = None,
        prefix: Optional[str] = None,
        op: Optional[str] = None,
    ) -> str:
        """
        Get the value for edge filter as defined by ``key``.
        This is used as a convenience method for generating cypher queries.

        Parameters
        ----------
        edge_filters: Dict
            All edge filters
        key: str
            Name of the edge filter
        variable: Optional[str]
            Variable binding for cypher query
        prefix: Optional[str]
            Prefix for the cypher
        op: Optional[str]
            The operator

        Returns
        -------
        str
            Value corresponding to the given edge filter ``key``, formatted for CQL

        """
        value = ""
        if key in edge_filters and edge_filters[key]:
            if isinstance(edge_filters[key], (list, set, tuple)):
                if key in {"subject_category", "object_category"}:
                    formatted = [f"{variable}{prefix}`{x}`" for x in edge_filters[key]]
                    value = f" {op} ".join(formatted)
                elif key == "predicate":
                    formatted = [f"'{x}'" for x in edge_filters["predicate"]]
                    value = f"type({variable}) IN [{', '.join(formatted)}]"
                elif key in knowledge_provenance_properties:
                    formatted = [
                        f"'{x}' IN {variable}{prefix}{key}" for x in edge_filters[key]
                    ]
                    value = f" {op} ".join(formatted)
                else:
                    formatted = []
                    for v in edge_filters[key]:
                        formatted.append(f"{variable}{prefix}{key} = '{v}'")
                    value = f" {op} ".join(formatted)
            elif isinstance(edge_filters[key], str):
                value = f"{variable}{prefix}{key} = '{edge_filters[key]}'"
            else:
                log.error(
                    f"Unexpected {key} edge filter of type {type(edge_filters[key])}"
                )
        return value
